import re
import warnings
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Literal, Type, Union

import pydantic as pyd
import pytest
from test_worker.main import EmptyStruct, IntStruct, MyGeneric

from tierkreis.builder import Const, Output, UnionConst, graph
from tierkreis.client.runtime_client import RuntimeClient
from tierkreis.core.opaque_model import OpaqueModel
from tierkreis.core.types import IncompatibleUnionType, TierkreisType, UnionTag
from tierkreis.core.values import (
    FloatValue,
    IntValue,
    StringValue,
    StructValue,
    TierkreisValue,
    ToPythonFailure,
    VariantValue,
    VecValue,
    option_none,
)


@pytest.mark.asyncio
async def test_union_types(bi, client: RuntimeClient) -> None:
    # test structs containing unions can be converted to variant values, sent
    # through worker functions that process them as unions, and converted back
    # to unions.
    iv = TierkreisValue.from_python(1)
    assert iv.to_python(float | int) == 1
    fv = TierkreisValue.from_python(2.3)
    assert fv.to_python(int | float) == 2.3

    @dataclass
    class StructWithUnion:
        # redefine struct with Union[] syntax here to test both
        # (alongside A | B syntax)
        x: Union[IntStruct, float]

    pn = bi["python_nodes"]
    a_py = StructWithUnion(1.0)
    b_py = StructWithUnion(IntStruct(2))

    @graph()
    def g() -> Output:
        return Output(
            a=pn.id_union_struct(Const(a_py)), b=pn.id_union_struct(Const(b_py))
        )

    out = await client.run_graph(g)
    a, b = out["a"], out["b"]

    assert isinstance(a, StructValue)
    assert isinstance(b, StructValue)
    a_var = a.values["x"]
    assert isinstance(a_var, VariantValue)
    assert a_var.tag == UnionTag.prefix + "float"
    b_var = b.values["x"]
    assert isinstance(b_var, VariantValue)
    assert b_var.tag == UnionTag.prefix + "int_struct"

    assert a.to_python(StructWithUnion) == a_py
    assert b.to_python(StructWithUnion) == b_py

    for bad_tag in ("invalid_tag", UnionTag.type_tag(IntStruct)):
        with pytest.raises(ToPythonFailure):
            v: VariantValue = VariantValue(bad_tag, TierkreisValue.from_python(1.0))
            struct: StructValue = StructValue({"x": v})
            _ = struct.to_python(StructWithUnion)

    @dataclass
    class AmbiguousUnion:
        x: Union[list[int], list[bool]]

    with pytest.raises(IncompatibleUnionType):
        TierkreisType.from_python(AmbiguousUnion)

    sig = await client.get_signature()

    @graph(type_check_sig=sig)
    def g2() -> Output:
        return Output(a=pn.id_union(UnionConst(1)), b=pn.id_union(UnionConst(1.0)))

    out = await client.run_graph(g2)
    a, b = out["a"], out["b"]
    assert a.to_python(int | float) == 1
    assert b.to_python(int | float) == 1.0


@pyd.dataclasses.dataclass(frozen=True, kw_only=True)
class ConstrainedField:
    foo: float
    bar: pyd.PositiveFloat | None = None
    points: tuple[float, ...]


@pyd.dataclasses.dataclass(frozen=True, kw_only=True)
class SimpleVariant(ConstrainedField):
    mean: float
    disc_name: Literal["Simple", "simple"] = field(default="Simple")


@pyd.dataclasses.dataclass(frozen=True, kw_only=True)
class ComplexVariant(ConstrainedField):
    mean: float
    variance: pyd.PositiveFloat
    foo: pyd.NonNegativeFloat
    disc_name: Literal["Complex"] = field(default="Complex", init=True)


class EnumExample(Enum):
    First = "First"
    Second = "Second"


class ContainsVariant(pyd.BaseModel):
    n: int
    enum: EnumExample
    variant: ComplexVariant | SimpleVariant = pyd.Field(discriminator="disc_name")


@pyd.dataclasses.dataclass(frozen=True, kw_only=True)
class NonInitPydField:
    xl: int = pyd.Field(default=0)
    non_init: int = field(default=0, init=False)


class Model(pyd.BaseModel):
    x: int
    y: NonInitPydField


class AnnotatedWithUnion(pyd.BaseModel):
    un: pyd.PositiveFloat | int


class AnnotatedInList(pyd.BaseModel):
    ul: list[pyd.PositiveFloat]


class OpaqueContainsVariant(OpaqueModel):
    n: int
    enum: EnumExample
    variant: ComplexVariant | SimpleVariant = pyd.Field(discriminator="disc_name")


@pytest.mark.asyncio
async def test_pydantic_types(bi, client: RuntimeClient) -> None:
    constrained = ConstrainedField(foo=1, points=(1.2,))
    tk_constrained = StructValue(
        {
            "foo": FloatValue(1.0),
            "points": VecValue([FloatValue(1.2)]),
            "bar": option_none,
        }
    )

    cmplex = ComplexVariant(mean=-3.4, variance=4.5, **constrained.__dict__)
    tk_cmplex = StructValue(
        {
            "mean": FloatValue(-3.4),
            "variance": FloatValue(4.5),
            "disc_name": StringValue("Complex"),
            **tk_constrained.values,
        }
    )

    simple = SimpleVariant(mean=1.3, **constrained.__dict__)
    tk_simple = StructValue(
        {
            **tk_constrained.values,
            "mean": FloatValue(1.3),
            "disc_name": StringValue("Simple"),
        }
    )

    contains = ContainsVariant(variant=cmplex, n=4, enum=EnumExample.First)
    opaque_contains = OpaqueContainsVariant(variant=cmplex, n=4, enum=EnumExample.First)

    tk_first = VariantValue("First", StructValue({}))
    tk_contains = StructValue(
        {
            "variant": VariantValue("Complex", tk_cmplex),
            "n": IntValue(4),
            "enum": tk_first,
        }
    )

    contains_alt = ContainsVariant(variant=simple, n=4, enum=EnumExample.Second)
    tk_second = VariantValue("Second", StructValue({}))
    tk_contains_alt = StructValue(
        {
            "variant": VariantValue("Simple", tk_simple),
            "n": IntValue(4),
            "enum": tk_second,
        }
    )

    da = NonInitPydField()
    tk_da = StructValue(
        {
            "xl": IntValue(0),
            "non_init": IntValue(0),
        }
    )

    tk_model = StructValue(
        {
            "x": IntValue(1),
            "y": tk_da,
        }
    )

    tk_list_model = StructValue(
        {
            "ul": VecValue([FloatValue(1.2)]),
        }
    )

    tk_union_model = StructValue(
        {
            "un": VariantValue(UnionTag.type_tag(int), IntValue(1)),
        }
    )
    tk_opaque_contains = StructValue(
        {
            "__tk_opaque_opaque_contains_variant": StringValue(
                opaque_contains.model_dump_json()
            ),
        }
    )

    def mk_mygeneric_variant(typ: Type, val: TierkreisValue) -> StructValue:
        return StructValue({"x": VariantValue(UnionTag.type_tag(typ), val)})

    samples: list[tuple[Type, Any, TierkreisValue]] = [
        (ConstrainedField, constrained, tk_constrained),
        (SimpleVariant, simple, tk_simple),
        (EnumExample, EnumExample.First, tk_first),
        (EnumExample, EnumExample.Second, tk_second),
        (ComplexVariant, cmplex, tk_cmplex),
        (ContainsVariant, contains, tk_contains),
        (ContainsVariant, contains_alt, tk_contains_alt),
        (NonInitPydField, da, tk_da),
        (Model, Model(x=1, y=da), tk_model),
        (AnnotatedInList, AnnotatedInList(ul=[1.2]), tk_list_model),
        (AnnotatedWithUnion, AnnotatedWithUnion(un=1), tk_union_model),
        (
            OpaqueContainsVariant,
            opaque_contains,
            tk_opaque_contains,
        ),
        (EnumExample | None, EnumExample.First, tk_first),
        # (None | EnumExample, EnumExample.First, tk_first), # fails, see test_reversed_union
        (EnumExample | None, None, option_none),
        (None | EnumExample, None, option_none),
    ]

    for py_type, py_val, expected_tk_val in samples:
        _ = TierkreisType.from_python(py_type)
        converted_tk_val = TierkreisValue.from_python(py_val)
        assert converted_tk_val == expected_tk_val
        assert TierkreisValue.from_python(py_val, py_type) == expected_tk_val
        assert converted_tk_val.to_python(py_type) == py_val

        pn = bi["python_nodes"]

        @graph()
        def g() -> Output:
            return Output(pn.id_py(Const(py_val)))

        out = (await client.run_graph(g))["value"]

        assert out.to_python(py_type) == py_val


def test_union_over_arguments() -> None:
    values = [MyGeneric[int](x=3), MyGeneric[str](x="foo")]
    types = [
        t
        for base in [MyGeneric[int] | MyGeneric[str], MyGeneric[str] | MyGeneric[int]]
        for t in [base, base | None, None | base]
    ]

    for val in values:
        for typ in types:
            tk_val = TierkreisValue.from_python(val, typ)
            assert tk_val.to_python(typ) == val

            # Separately test value+typ embedded together in an outer MyGeneric
            typ2 = MyGeneric[typ]
            val2 = typ2(x=val)
            tk_val2 = TierkreisValue.from_python(val2, typ2)
            assert tk_val2.to_python(typ2) == val2


@pytest.mark.xfail(reason="See issue #454")
def test_reversed_union_enum() -> None:
    # This should be an entry in test:pydantic_types, if it worked:
    # (None | EnumExample, EnumExample.First, tk_first)
    tk_first = VariantValue("First", StructValue({}))
    py_type, py_val, expected_tk_val = (None | EnumExample, EnumExample.First, tk_first)
    assert TierkreisValue.from_python(py_val) == expected_tk_val
    # Works the right way around:
    assert expected_tk_val.to_python(EnumExample | None) == py_val
    # But reverse the union and it does not (result is "None")
    assert expected_tk_val.to_python(py_type) == py_val


def test_pydantic_generic_union() -> None:
    # Similar to above but from_python with explicit 'typ'
    # avoids the need for pydantic validation
    val = MyGeneric(x=MyGeneric[float](x=3.2))
    typ = MyGeneric[MyGeneric[int] | MyGeneric[float]]
    tk_val = TierkreisValue.from_python(val, typ)
    assert tk_val.to_python(typ) == val


@pytest.mark.asyncio
async def test_return_union(bi, client: RuntimeClient) -> None:
    # test a worker function declared as returning a union type,
    # correctly converts the return value

    sig = await client.get_signature()
    pn = bi["python_nodes"]

    @graph(type_check_sig=sig)
    def g() -> Output:
        return Output(b=pn.zero_to_empty(x=Const(0)), a=pn.zero_to_empty(x=Const(1)))

    out = await client.run_graph(g)

    a, b = out["a"], out["b"]

    assert isinstance(a, VariantValue)
    assert a.tag == UnionTag.type_tag(IntStruct)
    assert isinstance(b, VariantValue)
    assert b.tag == UnionTag.type_tag(EmptyStruct)

    # to_python is tag-driven so ordering of union elements makes no difference
    assert a.to_python(IntStruct | EmptyStruct) == IntStruct(1)
    assert a.to_python(EmptyStruct | IntStruct) == IntStruct(1)
    assert b.to_python(IntStruct | EmptyStruct) == EmptyStruct()


# Make sure that the ordering of Union elements doesn't affect anything here
@pytest.mark.parametrize(
    "union_type", [IntStruct | EmptyStruct, EmptyStruct | IntStruct]
)
def test_warnings(union_type: Type) -> None:
    with warnings.catch_warnings(record=True) as ws:
        val = TierkreisValue.from_python(IntStruct(3), union_type)
    assert len(ws) == 0
    assert val == VariantValue(
        UnionTag.type_tag(IntStruct), StructValue({"y": IntValue(3)})
    )

    @dataclass
    class IntSubStruct(IntStruct):
        new_field: float

    with warnings.catch_warnings(record=True) as ws:
        val = TierkreisValue.from_python(IntSubStruct(13, 4.21), union_type)
    assert len(ws) == 1
    assert re.search("IntSubStruct.*beyond those in.*IntStruct", str(ws[0]))
    assert val == VariantValue(
        UnionTag.type_tag(IntStruct), StructValue({"y": IntValue(13)})
    )


@pytest.mark.asyncio
async def test_union_generic_basemodel(sig, bi, client: RuntimeClient) -> None:
    msg = "Hello World!"

    @graph(type_check_sig=sig)
    def main() -> Output:
        res = bi["python_nodes"].echo_union(Const(msg))
        return Output(res)

    res = await client.run_graph(main)
    assert res.pop("value").to_python(MyGeneric[str] | int) == MyGeneric[str](x=msg)
    assert res == {}  # No other outputs besides x



import pytest
from sample_graph import sample_graph as sample_g

from tierkreis.core.tierkreis_graph import TierkreisEdge, TierkreisGraph
from tierkreis.core.values import TierkreisValue, VariantValue
from tierkreis.pyruntime import PyRuntime


@pytest.fixture()
def sample_graph() -> TierkreisGraph:
    return sample_g()


@pytest.mark.asyncio
async def test_callback(sample_graph: TierkreisGraph, pyruntime_function: PyRuntime):
    ins = {"inp": "world", "vv": VariantValue("many", TierkreisValue.from_python(2))}

    cache = {}

    def cback(e: TierkreisEdge, v: TierkreisValue):
        cache[e] = v

    pyruntime_function.set_callback(cback)
    outs = await pyruntime_function.run_graph(sample_graph, **ins)
    assert all(e in cache for e in sample_graph.edges())
    assert sorted(outs) == sorted(sample_graph.outputs())



"""Simple utilities useful throughout the codebase."""

from typing import Callable, Mapping, Optional, TypeVar, Union

from tierkreis.core.function import FunctionDeclaration, FunctionName
from tierkreis.core.tierkreis_graph import TierkreisGraph

K = TypeVar("K")
V = TypeVar("V")
W = TypeVar("W")


def map_vals(inp: Mapping[K, V], f: Callable[[V], W]) -> dict[K, W]:
    """Apply function `f` to map all the values in the input mapping."""
    return {k: f(v) for k, v in inp.items()}


def rename_ports_graph(name_map: dict[str, str]) -> TierkreisGraph:
    """Generate an identity graph with inputs -> outputs mapped by the provided
    dictionary.
    """
    tg = TierkreisGraph()

    tg.set_outputs(**{out: tg.input[inp] for inp, out in name_map.items()})

    return tg


def graph_from_func(
    name: Union[str, FunctionName],
    func: FunctionDeclaration,
    input_map: Optional[dict[str, str]] = None,
    output_map: Optional[dict[str, str]] = None,
) -> TierkreisGraph:
    """Build a graph with a single function node.
    If (input_map/output_map) is provided keys of the map which are
    (inputs/outputs) to the function
    are connected to the value as the (input/output) to the graph.
    For any missing keys, the same port name as the function is used.
    """
    inmap = input_map or {}
    oumap = output_map or {}

    for mp, order in [(inmap, func.input_order), (oumap, func.output_order)]:
        for port in order:
            if port not in mp:
                mp[port] = port

    tg = TierkreisGraph()
    node = tg.add_func(name, **{k: tg.input[v] for (k, v) in inmap.items()})
    tg.set_outputs(**{v: node[k] for k, v in oumap.items()})
    return tg



"""Implementation of simple python-only runtime."""

import asyncio
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Callable, Iterable, Optional, Tuple, cast

import networkx as nx
import requests

from tierkreis.client.runtime_client import RuntimeClient
from tierkreis.core import Labels
from tierkreis.core.function import FunctionName
from tierkreis.core.protos.tierkreis.v1alpha1.graph import Output, OutputStream
from tierkreis.core.signature import Signature
from tierkreis.core.tierkreis_graph import (
    BoxNode,
    ConstNode,
    FunctionNode,
    GraphValue,
    IncomingWireType,
    InputNode,
    MatchNode,
    OutputNode,
    TagNode,
    TierkreisEdge,
    TierkreisGraph,
)
from tierkreis.core.type_errors import TierkreisTypeErrors
from tierkreis.core.type_inference import _TYPE_CHECK, infer_graph_types
from tierkreis.core.utils import map_vals
from tierkreis.core.values import StructValue, TierkreisValue, VariantValue, VecValue
from tierkreis.pyruntime import python_builtin

if TYPE_CHECKING:
    from tierkreis.core.tierkreis_graph import _EdgeData
    from tierkreis.worker.namespace import Namespace


class _ValueNotFound(Exception):
    def __init__(self, tke: TierkreisEdge) -> None:
        self.edge = tke
        super().__init__(f"Value not found on edge {tke.source} -> {tke.target}")


class OutputNotFound(_ValueNotFound):
    """Node output expected but not found."""


class InputNotFound(_ValueNotFound):
    """Node input expected but not found."""


class FunctionNotFound(Exception):
    """Function expected but not found."""

    def __init__(self, fname: FunctionName) -> None:
        self.function = fname
        super().__init__(f"Function {fname} not found in namespace.")


class PyRuntime(RuntimeClient):
    """A simplified python-only Tierkreis runtime. Can be used with builtin
    operations and python only namespaces that are locally available."""

    def __init__(self, roots: Iterable["Namespace"], num_workers: int = 1):
        """Initialise with locally available namespaces, and the number of
        workers (asyncio tasks) to use in execution."""
        self.root = deepcopy(python_builtin.namespace)
        for root in roots:
            self.root.merge_namespace(root)
        self.num_workers = num_workers
        self._callback: Optional[Callable[[TierkreisEdge, TierkreisValue], None]] = None
        self.set_callback(None)

    def set_callback(
        self, callback: Optional[Callable[[TierkreisEdge, TierkreisValue], None]]
    ):
        """Set a callback function that takes a TierkreisEdge and
        TierkreisValue, which will be called every time a edge receives an
        output. Can be used to inspect intermediate values."""
        self._callback = callback

    def callback(
        self,
        edge: TierkreisEdge,
        val: TierkreisValue,
    ):
        """If a callback function is set, call it with an edge and the value on
        the edge."""
        if self._callback:
            self._callback(edge, val)

    async def run_graph(
        self,
        run_g: TierkreisGraph,
        /,
        **py_inputs: Any,
    ) -> dict[str, TierkreisValue]:
        """Run a tierkreis graph using the python runtime, and provided inputs.
        Returns the outputs of the graph.
        """
        total_nodes = run_g._graph.number_of_nodes()
        runtime_state: dict["_EdgeData", TierkreisValue] = {}

        async def run_node(node: int) -> dict[str, TierkreisValue]:
            tk_node = run_g[node]

            if isinstance(tk_node, OutputNode):
                return {}
            if isinstance(tk_node, InputNode):
                return map_vals(py_inputs, TierkreisValue.from_python)

            if isinstance(tk_node, ConstNode):
                return {Labels.VALUE: tk_node.value}

            in_edges = list(run_g.in_edges(node))
            while not all(e.to_edge_handle() in runtime_state for e in in_edges):
                # wait for inputs to become available
                # only useful if there are other workers that can do things
                # while this one waits
                assert self.num_workers > 1
                await asyncio.sleep(0)
            try:
                in_values = (
                    (e, runtime_state.pop(e.to_edge_handle())) for e in in_edges
                )
            except KeyError as key_e:
                raise InputNotFound(run_g._to_tkedge(key_e.args[0])) from key_e

            inps = {e.target.port: val for e, val in in_values}
            if isinstance(tk_node, FunctionNode):
                fname = tk_node.function_name
                if fname.namespaces == [] and fname.name == "eval":
                    return await self._run_eval(inps)
                elif fname.namespaces == [] and fname.name == "loop":
                    return await self._run_loop(inps)
                elif fname.namespaces == [] and fname.name == "map":
                    return await self._run_map(inps)
                else:
                    function = self.root.get_function(fname)
                    if function is None:
                        raise FunctionNotFound(fname)
                    # For now the PyRuntime does not provide a stack trace
                    return (await function.run(self, dict(), StructValue(inps))).values

            elif isinstance(tk_node, BoxNode):
                return await self.run_graph(
                    tk_node.graph,
                    **inps,
                )

            elif isinstance(tk_node, MatchNode):
                return self._run_match(inps)

            elif isinstance(tk_node, TagNode):
                return {
                    Labels.VALUE: VariantValue(tk_node.tag_name, inps[Labels.VALUE])
                }

            else:
                raise RuntimeError("Unknown node type.")

        async def worker(queue: asyncio.Queue[int]):
            # each worker gets the next node in the queue
            while True:
                node = await queue.get()
                # If the node is not yet runnable,
                # wait/block until it is, do not try to run any other node
                outs = await run_node(node)

                # assign outputs to edges
                for out_edge in run_g.out_edges(node):
                    try:
                        val = outs.pop(out_edge.source.port)
                    except KeyError as key_e:
                        raise OutputNotFound(out_edge) from key_e
                    tkval = TierkreisValue.from_python(val)
                    self.callback(out_edge, tkval)
                    runtime_state[out_edge.to_edge_handle()] = tkval
                # signal this node is now done
                queue.task_done()

        que: asyncio.Queue[int] = asyncio.Queue(total_nodes)
        for node in nx.topological_sort(run_g._graph):
            # add all node names to the queue in topsort order
            # if there are fewer workers than nodes, and the queue is populated
            # in a non-topsort order, some worker may just wait forever for it's
            # node's inputs to become available.
            que.put_nowait(node)

        workers = [asyncio.create_task(worker(que)) for _ in range(self.num_workers)]
        queue_complete = asyncio.create_task(que.join())

        # wait for either all nodes to complete, or for a worker to return
        await asyncio.wait(
            [queue_complete, *workers], return_when=asyncio.FIRST_COMPLETED
        )
        if not queue_complete.done():
            # If the queue hasn't completed, it means one of the workers has
            # raised - find it and propagate the exception.
            # even if the rest of the graph has not completed
            for t in workers:
                if t.done():
                    t.result()  # this will raise
        for task in workers:
            task.cancel()

        # Wait until all worker tasks are cancelled.
        await asyncio.gather(*workers, return_exceptions=True)

        return {
            e.target.port: runtime_state.pop(e.to_edge_handle())
            for e in run_g.in_edges(run_g.output_node_idx)
        }

    async def _run_eval(
        self, ins: dict[str, TierkreisValue]
    ) -> dict[str, TierkreisValue]:
        thunk = cast(GraphValue, ins.pop(Labels.THUNK)).value
        return await self.run_graph(thunk, **ins)

    async def _run_loop(
        self, ins: dict[str, TierkreisValue]
    ) -> dict[str, TierkreisValue]:
        body = cast(GraphValue, ins.pop("body")).value
        while True:
            outs = await self.run_graph(
                body,
                **ins,
            )
            out = cast(
                VariantValue,
                outs[Labels.VALUE],
            )
            nxt = {"value": out.value}
            if out.tag == Labels.BREAK:
                return nxt
            else:
                ins = nxt

    async def _run_map(
        self, ins: dict[str, TierkreisValue]
    ) -> dict[str, TierkreisValue]:
        body = cast(GraphValue, ins.pop("thunk")).value
        inputs = cast(VecValue, ins.pop("value")).values

        async def task(x):
            return (await self.run_graph(body, value=x))[Labels.VALUE]

        tasks = [asyncio.create_task(task(x)) for x in inputs]
        out = await asyncio.gather(*tasks)
        ret = {"value": cast(TierkreisValue, VecValue(out))}
        return ret

    def _run_match(self, ins: dict[str, TierkreisValue]) -> dict[str, TierkreisValue]:
        variant = cast(VariantValue, ins[Labels.VARIANT_VALUE])
        thunk = cast(GraphValue, ins[variant.tag]).value

        newg = TierkreisGraph()
        boxinps: dict[str, IncomingWireType] = {
            inp: newg.input[inp] for inp in thunk.inputs()
        }
        boxinps[Labels.VALUE] = newg.add_const(variant.value)
        box = newg.add_box(thunk, **boxinps)
        newg.set_outputs(**{out: box[out] for out in thunk.outputs()})
        return {Labels.THUNK: GraphValue(newg)}

    async def get_signature(self) -> Signature:
        return self.root.extract_signature(True)

    async def type_check_graph(self, graph: TierkreisGraph) -> TierkreisGraph:
        return infer_graph_types(graph, await self.get_signature())

    async def type_check_graph_with_inputs(
        self, tg, inputs: StructValue
    ) -> Tuple[TierkreisGraph, StructValue]:
        return infer_graph_types(tg, await self.get_signature(), inputs)

    @property
    def can_type_check(self) -> bool:
        return _TYPE_CHECK


class VizRuntime(PyRuntime):
    """Child class of ``PyRuntime`` that can interact with a tierkreis-viz instance
    for live graph visulization."""

    def __init__(self, url: str, roots: Iterable["Namespace"], num_workers: int = 1):
        """`url` is the address of the running tierkreis-viz instance. See
        `PyRuntime` for remaining parameters
        """
        self.url = url
        self.outputs = OutputStream()
        super().__init__(roots, num_workers)

    def _post(self, endpoint: str, data):
        proto_dat = data.to_proto() if hasattr(data, "to_proto") else data
        requests.post(
            self.url + endpoint,
            data=bytes(proto_dat),
            headers={"content-type": "application/protobuf"},
        )

    async def type_check_graph(self, graph: TierkreisGraph) -> TierkreisGraph:
        """See ``PyRuntime.type_check_graph``. Additionally updates
        vizualized graph with type annotations."""
        try:
            typedg = await super().type_check_graph(graph)
        except TierkreisTypeErrors as e:
            self._post("/api/typeErrors", e)
            raise e
        self.viz_graph(typedg)
        return typedg

    def viz_graph(self, tg: TierkreisGraph):
        """Send graph to be visualized."""
        self._post("/api/graph", tg)
        self._post("/api/streamList", OutputStream())
        self._post("/api/typeErrors", TierkreisTypeErrors([]))

    async def run_viz_graph(
        self,
        run_g: TierkreisGraph,
        /,
        **py_inputs: Any,
    ) -> dict[str, TierkreisValue]:
        """See ``PyRuntime.run_graph``. Additionally updates the
        visualization with the outputs of each node when they are available."""
        self.outputs = OutputStream()
        return await self.run_graph(run_g, **py_inputs)

    def callback(
        self,
        edge: TierkreisEdge,
        val: TierkreisValue,
    ):
        super().callback(edge, val)
        self.outputs.stream.append(Output(edge=edge.to_proto(), value=val.to_proto()))
        self._post("/api/streamList", self.outputs)



import glob
import os
import subprocess

from setuptools import build_meta as default_backend  # type : ignore

# PEP 517 / 518 local (aka in-tree) backend for pip/build/etc.
# We expect to be run from the project root, i.e. python/
PROTO_INPUT_DIR = "./protos"
PROTO_OUTPUT_DIR = "./tierkreis/core/protos"


def generate_proto_code():
    proto_files = glob.glob(PROTO_INPUT_DIR + "/**/*.proto", recursive=True)
    assert len(proto_files) > 0

    if not os.path.exists(PROTO_OUTPUT_DIR):
        os.mkdir(PROTO_OUTPUT_DIR)

    subprocess.check_call(
        [
            "protoc",
            "--python_betterproto_out=" + PROTO_OUTPUT_DIR,
            "--experimental_allow_proto3_optional",
            "-I",
            PROTO_INPUT_DIR,
        ]
        + proto_files
    )


def build_sdist(*args, **kwargs):
    generate_proto_code()
    return default_backend.build_sdist(*args, **kwargs)


def build_wheel(*args, **kwargs):
    if os.path.exists(PROTO_INPUT_DIR):
        generate_proto_code()
    else:
        # Building wheel from sdist.
        # Protos should have been generated already
        # and placed into the sdist.
        proto_files_already_generated = glob.glob(
            PROTO_OUTPUT_DIR + "/tierkreis/**/*.py", recursive=True
        )
        assert len(proto_files_already_generated) > 0
        # No generation required.
    return default_backend.build_wheel(*args, **kwargs)


get_requires_for_build_sdist = default_backend.get_requires_for_build_sdist

get_requires_for_build_wheel = default_backend.get_requires_for_build_wheel

prepare_metadata_for_build_wheel = default_backend.prepare_metadata_for_build_wheel

# PEP 660 defines additional hooks to support "pip install --editable":
def build_editable(*args, **kwargs):
    generate_proto_code()
    return default_backend.build_editable(*args, **kwargs)


get_requires_for_build_editable = default_backend.get_requires_for_build_editable

prepare_metadata_for_build_editable = (
    default_backend.prepare_metadata_for_build_editable
)



from __future__ import annotations

import base64
import json
import zlib
from abc import ABC, abstractmethod
from dataclasses import dataclass
from importlib.util import find_spec
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Iterable,
    List,
    Protocol,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from tierkreis.core.opaque_model import OpaqueModel

if TYPE_CHECKING:
    from pytket._tket.circuit import Circuit
    from pytket.backends.backendresult import BackendResult
    from pytket.circuit import Node
    from typing_extensions import Self


_PYTKET = find_spec("pytket") is not None


def _pytket():
    if not _PYTKET:
        raise PytketDependencyError
    import pytket

    return pytket


class PytketDependencyError(Exception):
    """Exception for pytket import error"""

    def __init__(
        self,
        msg="pytket is not installed. "
        "Install optional feature tierkreis['common_types']",
        *args,
        **kwargs,
    ) -> None:
        super().__init__(msg, *args, **kwargs)


@dataclass(frozen=True)
class UnitID:
    """Used for BackendResult.{qubits, bits} and also for Node.
    Not serializable on its own. Returns / takes a list.
    """

    reg_name: str
    index: List[int]

    def to_serializable(self) -> Tuple[str, List[int]]:
        return self.reg_name, self.index

    @classmethod
    def from_serializable(cls, jsonable: Iterable[Union[str, List[int]]]) -> "UnitID":
        reg_name, index = jsonable
        return UnitID(reg_name=cast(str, reg_name), index=cast(List[int], index))

    @classmethod
    def from_pytket_node(cls, node: Node) -> "UnitID":
        """Convert a pytket Node object to a UnitID object."""
        return UnitID.from_serializable(node.to_list())

    def to_pytket_node(self) -> Node:
        """Convert a UnitID object to a pytket Node object."""
        _pytket().circuit.Node  # type: ignore

        return Node.from_list(list(self.to_serializable()))

    def __hash__(self) -> int:
        return hash(str(self))

    def __eq__(self, other: Any) -> bool:
        return str(self) == str(other)


class PytketType(Protocol):
    """Many pytket types define to_dict and from_dict for serialization."""

    def to_dict(self) -> dict[str, Any]: ...

    @classmethod
    def from_dict(cls, d: dict[str, Any], /) -> "PytketType": ...


T = TypeVar("T", bound=PytketType)


def json_zip(j: dict[str, Any]) -> str:
    return base64.b64encode(zlib.compress(json.dumps(j).encode("utf-8"))).decode(
        "ascii"
    )


def json_unzip(j: str) -> dict[str, Any]:
    return json.loads(zlib.decompress(base64.b64decode(j)))


class PytketWrapper(OpaqueModel, ABC, Generic[T]):
    """Generic wrapper around pytket types with to_dict and from_dict methods."""

    json_str: str

    @classmethod
    @abstractmethod
    def get_pytket_type(cls) -> Type[T]:
        raise NotImplementedError

    @classmethod
    def new_tierkreis_compatible(cls, value: T) -> Self:
        return cls(json_str=json_zip(value.to_dict()))

    def from_tierkreis_compatible(self) -> T:
        return self.get_pytket_type().from_dict(json_unzip(self.json_str))


class CircuitWrapper(PytketWrapper["Circuit"]):
    """Tierkreis wrapper for pytket Circuit."""

    @classmethod
    def get_pytket_type(cls) -> Type["Circuit"]:
        return _pytket().circuit.Circuit  # type: ignore


class ResultWrapper(PytketWrapper["BackendResult"]):
    """Tierkreis wrapper for pytket BackendResult."""

    @classmethod
    def get_pytket_type(cls) -> Type["BackendResult"]:
        return _pytket().backends.backendresult.BackendResult  # type: ignore


def register_pytket_types():
    """Register conversions for pytket types with Tierkreis."""
    if _PYTKET:
        from pytket._tket.circuit import Circuit
        from pytket.backends.backendresult import BackendResult

        from tierkreis.core.types import TierkreisType

        TierkreisType.register_alternative(Circuit, CircuitWrapper)
        TierkreisType.register_alternative(BackendResult, ResultWrapper)
    else:
        raise PytketDependencyError



from abc import ABC, abstractmethod
from typing import Any, Dict

from tierkreis.core.signature import Signature
from tierkreis.core.tierkreis_graph import TierkreisGraph
from tierkreis.core.values import TierkreisValue


class RuntimeClient(ABC):
    """Abstract class for clients which can run tierkreis graphs on a runtime."""

    @abstractmethod
    async def get_signature(self) -> Signature:
        """Get the signature of functions available on the runtime."""

    @abstractmethod
    async def run_graph(
        self,
        graph: TierkreisGraph,
        /,
        **py_inputs: Any,
    ) -> Dict[str, TierkreisValue]:
        """Execute a tierkreis graph on the runtime. Inputs are taken as keyword arguments and each keyword argument must match an input port of the graph.  Non-tierkreis values will be converted to tierkreis values if possible."""

    @abstractmethod
    async def type_check_graph(self, graph: TierkreisGraph) -> TierkreisGraph:
        """Type check a tierkreis graph on the runtime against the runtime signature."""

    @property
    def can_type_check(self) -> bool:
        """Whether the runtime supports type-checking graphs."""
        return True



"""Namespace class for holding namespace definitions of python Tierkreis worker."""

import dataclasses
import typing
from ctypes import ArgumentError
from dataclasses import dataclass, make_dataclass
from functools import wraps
from inspect import getdoc, isclass
from typing import (
    Awaitable,
    Callable,
    Dict,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Type,
    Union,
    cast,
)

from tierkreis.client.runtime_client import RuntimeClient
from tierkreis.core.function import FunctionDeclaration, FunctionName
from tierkreis.core.signature import Namespace as SigNamespace
from tierkreis.core.signature import Signature
from tierkreis.core.tierkreis_graph import Location
from tierkreis.core.types import (
    Constraint,
    GraphType,
    Kind,
    Row,
    StarKind,
    StructType,
    TierkreisType,
    TypeScheme,
    UnpackRow,
)
from tierkreis.core.values import (
    StructValue,
    TierkreisValue,
)
from tierkreis.worker.exceptions import (
    DecodeInputError,
    EncodeOutputError,
    NamespaceClash,
    NodeExecutionError,
)

from .tracing import get_tracer, span

tracer = get_tracer(__name__)

Metadata = MutableMapping[str, str | bytes]

METADATA_ARG = "tierkreis_metadata"
CALLBACK_ARG = "client"


@dataclass
class Function:
    """Registered python function defined by a callable and a tierkreis function declaration."""

    run: Callable[[RuntimeClient, Metadata, StructValue], Awaitable[StructValue]]
    declaration: FunctionDeclaration


def _snake_to_pascal(name: str) -> str:
    return name.replace("_", " ").title().replace(" ", "")


def _get_base_tkstruct(class_type: Type) -> Type:
    origin = cast(Type, typing.get_origin(class_type))
    return origin if origin is not None else class_type


def _get_ordered_names(struct: Type) -> List[str]:
    tk_cls = _get_base_tkstruct(struct)
    return [field.name for field in dataclasses.fields(tk_cls)]


# Convert the type vars to names
def _type_var_to_name(type_var: Union[str, typing.TypeVar]) -> str:
    if isinstance(type_var, typing.TypeVar):
        return type_var.__name__
    return type_var


def _check_tkstruct_hint(hint: Type) -> bool:
    tk_cls = _get_base_tkstruct(hint)
    return tk_cls is not None and isclass(tk_cls) and issubclass(tk_cls, UnpackRow)


class Namespace(Mapping[str, "Namespace"]):
    """Namespace containing Tierkreis Functions, keyed by function name.
    Used to construct Tierkreis namespaces from python defined functions in workers.
    """

    functions: dict[str, Function]
    aliases: dict[str, TypeScheme]
    subspaces: dict[str, "Namespace"]

    def __init__(self):
        self.functions = {}
        self.aliases = {}
        self.subspaces = {}

    def __getitem__(self, __k: str) -> "Namespace":
        return self.subspaces.setdefault(__k, Namespace())

    def __iter__(self):
        return self.subspaces.__iter__()

    def __len__(self):
        return self.subspaces.__len__()

    def merge_namespace(self, other: "Namespace"):
        """Merge two namespaces together, joining on keys.
        On collisions raises :class:`~tierkreis.worker.exceptions.NamespaceClash`."""
        self._merge_namespace(other, [])

    def _merge_namespace(self, other: "Namespace", prefix: list[str]):
        """Merges other namespace into self"""
        intersect = self.functions.keys() & other.functions.keys()
        if intersect:
            raise NamespaceClash(prefix, intersect)
        self.functions = other.functions | self.functions
        for k, v in other.subspaces.items():
            if (x := self.subspaces.get(k)) is None:
                self.subspaces[k] = v
            else:
                x._merge_namespace(v, prefix + ["k"])

    def add_alias(self, name: str, type_: Type) -> Type:
        """Add a type alias to the namespace."""
        self.aliases[name] = TypeScheme({}, [], TierkreisType.from_python(type_))
        return type_

    def add_named_struct(self, name, type_: Type) -> Type:
        """Add a named struct to the namespace."""
        tk_type = TierkreisType.from_python(type_)
        if not isinstance(tk_type, StructType):
            raise ValueError(f"{type_} cannot be converted to a Tierkreis Struct Type.")
        tk_type.name = name
        self.aliases[name] = TypeScheme({}, [], tk_type)
        return type_

    def extract_contents(self) -> SigNamespace:
        """Convert to a generic namespace used to define Tierkreis signatures."""
        return SigNamespace(
            functions={k: v.declaration for k, v in self.functions.items()},
            subspaces={k: v.extract_contents() for k, v in self.subspaces.items()},
        )

    def extract_aliases(self) -> dict[str, TypeScheme]:
        """Get a mapping from type alias names to type schemes."""
        return self.aliases | {
            f"{name}::{k}": v
            for name, ns in self.subspaces.items()
            for k, v in ns.extract_aliases().items()
        }

    def extract_signature(self, can_scope: bool) -> Signature:
        """Convert to a Tierkreis signature."""
        return Signature(
            root=self.extract_contents(),
            aliases=self.extract_aliases(),
            scopes=[Location([])] if can_scope else [],
        )

    def get_function(self, name: FunctionName) -> Optional[Function]:
        """Recursively search namespace tree for function by name."""
        ns = self
        for x in name.namespaces:
            if (subns := ns.subspaces.get(x)) is None:
                return None
            ns = subns
        return ns.functions.get(name.name)

    def function(
        self,
        name: Optional[str] = None,
        constraints: Optional[List[Constraint]] = None,
        type_vars: Optional[Dict[Union[str, typing.TypeVar], Kind]] = None,
        callback: bool = False,
        metadata_keys: list[str] | None = None,
    ) -> Callable[[Callable], Callable]:
        """Decorator to register a python function as a Tierkreis function
        within the namespace.

        Args:
            name: Optionally explicitly set
                the name of the function, defaults to None (in which
                case the function name is used)
            constraints: Type constraints needed to define the type scheme of the
                function.
            type_vars:
                Declare type variables used in the function signature.
            callback: Whether the function expects a
                callback client as the first argument, in order to make
                graph execution requests.
            metadata_keys: List of keys in
                the metadata dictionary that the function uses. If
                present, the function should expect a parameter named
                `tierkreis_metadata` which is a dictionary with
                specified keys, mapped to the values present in the
                function request GRPC metadata.
        """

        def decorator(func: Callable) -> Callable:
            func_name = name or func.__name__

            # Get input and output type hints
            type_hints = typing.get_type_hints(func)

            if "return" not in type_hints:
                raise ValueError("Tierkreis function needs return type hint.")
            return_hint = type_hints.pop("return")

            struct_input = "inputs" in type_hints and _check_tkstruct_hint(
                type_hints["inputs"]
            )

            if callback:
                try:
                    type_hints.pop(CALLBACK_ARG)
                except KeyError:
                    raise ArgumentError(
                        "Functions with callbacks must have an argument 'client'"
                    )

            if metadata_keys is not None:
                try:
                    type_hints.pop(METADATA_ARG)
                except KeyError:
                    raise ArgumentError(
                        f"Functions asking for metadata values must have an argument '{METADATA_ARG}'"
                    )

            hint_inputs: Type = (
                type_hints["inputs"]
                if struct_input
                else make_dataclass(
                    f"{_snake_to_pascal(func_name)}Inputs", type_hints.items()
                )
            )

            struct_output = _check_tkstruct_hint(return_hint)
            hint_outputs: Type = (
                return_hint
                if struct_output
                else make_dataclass(
                    f"{_snake_to_pascal(func_name)}Outputs", [("value", return_hint)]
                )
            )

            # Wrap function with input and output conversions
            @wraps(func)
            async def wrapped_func(
                runtime: RuntimeClient,
                metadata: Metadata,
                inputs: StructValue,
            ) -> StructValue:
                try:
                    with span(tracer, name="decoding inputs to python type"):
                        python_inputs = (
                            {"inputs": inputs.to_python(hint_inputs)}
                            if struct_input
                            else {
                                name: val.to_python(type_hints[name])
                                for name, val in inputs.values.items()
                            }
                        )
                except Exception as error:
                    raise DecodeInputError(str(error)) from error

                if callback:
                    python_inputs[CALLBACK_ARG] = runtime
                if metadata_keys is not None:
                    # unavailable keys are ignored
                    python_inputs[METADATA_ARG] = {
                        k: metadata.pop(k) for k in metadata_keys if k in metadata
                    }
                try:
                    python_outputs = await func(**python_inputs)
                except Exception as error:
                    raise NodeExecutionError(error) from error

                try:
                    with span(tracer, name="encoding outputs from python type"):
                        return_type = hint_outputs if struct_output else return_hint
                        outputs = TierkreisValue.from_python(
                            python_outputs, return_type
                        )

                except Exception as error:
                    raise EncodeOutputError(str(error)) from error
                return (
                    cast(StructValue, outputs)
                    if struct_output
                    else StructValue({"value": outputs})
                )

            type_vars_by_name = (
                {_type_var_to_name(var): kind for var, kind in type_vars.items()}
                if type_vars
                else {}
            )

            # Convert type hints into tierkreis types
            type_inputs = Row.from_python(hint_inputs)
            type_outputs = Row.from_python(hint_outputs)
            type_vars_by_name.update(
                {name: StarKind() for name in type_inputs.contained_vartypes()}
            )
            type_vars_by_name.update(
                {name: StarKind() for name in type_outputs.contained_vartypes()}
            )

            # Construct the type schema of the function
            type_scheme = TypeScheme(
                constraints=constraints or [],
                variables=type_vars_by_name,
                body=GraphType(
                    inputs=type_inputs,
                    outputs=type_outputs,
                ),
            )

            self.functions[func_name] = Function(
                run=wrapped_func,
                declaration=FunctionDeclaration(
                    type_scheme=type_scheme.to_proto(),
                    description=getdoc(func) or "",
                    input_order=_get_ordered_names(hint_inputs),
                    output_order=_get_ordered_names(hint_outputs),
                ),
            )
            return func

        return decorator



from __future__ import annotations

import typing
from dataclasses import dataclass
from typing import Generic

import tierkreis.core.protos.tierkreis.v1alpha1.graph as pg
from tierkreis.core.types import UnpackRow

if typing.TYPE_CHECKING:
    from tierkreis.core.tierkreis_graph import TierkreisGraph


In = typing.TypeVar("In", bound=UnpackRow)
Out = typing.TypeVar("Out", bound=UnpackRow)


@dataclass
class RuntimeGraph(Generic[In, Out]):
    "Graph with a `RuntimeStruct` annotation for inputs and outputs."

    graph: TierkreisGraph

    def to_proto(self) -> pg.Value:
        return pg.Value(graph=self.graph.to_proto())



"""Callback client."""

from contextlib import asynccontextmanager
from typing import Any, AsyncIterator, Dict
from urllib.parse import urlparse

from grpclib.client import Channel

import tierkreis.core.protos.tierkreis.v1alpha1.runtime as pr
from tierkreis.client.runtime_client import RuntimeClient
from tierkreis.client.server_client import ServerRuntime
from tierkreis.core.signature import Signature
from tierkreis.core.tierkreis_graph import Location, TierkreisGraph
from tierkreis.core.values import TierkreisValue


class Callback(RuntimeClient):
    """Runtime client for use within a worker function, to submit graphs back to
    the parent runtime.
    """

    def __init__(self, channel: Channel, loc: Location):
        self.runtime = ServerRuntime(channel)
        self.loc = loc

    async def get_signature(self) -> Signature:
        return await self.runtime.get_signature(self.loc)

    async def run_graph(
        self,
        graph: TierkreisGraph,
        /,
        **py_inputs: Any,
    ) -> Dict[str, TierkreisValue]:
        return await self.runtime.run_graph(graph, self.loc, **py_inputs)

    async def type_check_graph(self, graph: TierkreisGraph) -> TierkreisGraph:
        return await self.runtime.type_check_graph(graph, self.loc)


@asynccontextmanager
async def callback_server(callback: pr.Callback) -> AsyncIterator[RuntimeClient]:
    """Context manager for connection to a callback server."""
    url = urlparse(callback.uri)
    host, port = url.hostname, url.port
    assert host is not None
    async with Channel(host, port) as channel:
        yield Callback(channel, callback.loc)



"""The builder module allows Tierkreis graphs to be constructed concisely using
decorated functions and context managers.
"""

import inspect
from abc import ABC, abstractmethod
from contextlib import AbstractContextManager
from contextvars import ContextVar, Token
from dataclasses import dataclass
from functools import cached_property
from itertools import count
from types import FrameType, TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from yachalk import chalk

from tierkreis.core import Labels
from tierkreis.core.function import FunctionDeclaration, FunctionName
from tierkreis.core.signature import Namespace as SigNamespace
from tierkreis.core.signature import Signature
from tierkreis.core.tierkreis_graph import (
    BoxNode,
    ConstNode,
    FunctionNode,
    IncomingWireType,
    Location,
    MatchNode,
    NodePort,
    NodeRef,
    PortID,
    TagNode,
    TierkreisGraph,
    TierkreisNode,
    to_nodeport,
)
from tierkreis.core.type_errors import TierkreisTypeErrors
from tierkreis.core.type_inference import infer_graph_types
from tierkreis.core.types import TupleLabel, TypeScheme, UnionTag
from tierkreis.core.utils import graph_from_func, map_vals, rename_ports_graph
from tierkreis.core.values import (
    TierkreisValue,
)

if TYPE_CHECKING:
    from tierkreis.core.types import GraphType, TierkreisType


class PortFunc(ABC):
    """Magic for lazily creating nodes (and returning ports thereof)."""

    @abstractmethod
    def resolve(self) -> NodePort:
        """Return a NodePort, creating nodes as necessary in source graph."""

    @abstractmethod
    def source_graph(self) -> TierkreisGraph:
        """The graph the port belongs to (or will)."""


ValueSource = Union[IncomingWireType, PortFunc]


class StablePortFunc(PortFunc):
    """Extends the contract of PortFunc, as documented for the resolve method"""

    @abstractmethod
    def resolve(self) -> NodePort:
        """As superclass, but guarantees that all calls return the same NodePort."""

    def __getitem__(self, field_name: str) -> "Unpack":
        # Syntactic sugar for unpacking structs
        return Unpack(self, field_name)


def _resolve(s: ValueSource) -> NodePort:
    return s.resolve() if isinstance(s, PortFunc) else to_nodeport(s)


def _source_graph(s: ValueSource) -> TierkreisGraph:
    return (
        s.source_graph() if isinstance(s, PortFunc) else to_nodeport(s).node_ref.graph
    )


def _debug_str() -> str:
    # unroll current stack until out of this file
    frm = cast(FrameType, inspect.currentframe())
    while frm.f_code.co_filename == __file__ and frm.f_back is not None:
        frm = cast(FrameType, frm.f_back)
    try:
        return f"Node added at: {frm.f_code.co_filename}:{frm.f_lineno}"
    finally:
        # as recommended in
        # https://docs.python.org/3/library/inspect.html#the-interpreter-stack
        del frm


__state: ContextVar["GraphBuilder"] = ContextVar("state")


class InvalidContext(Exception):
    """Expression not valid outside GraphBuilder `with` context"""


def current_builder() -> "GraphBuilder":
    """The current GraphBuilder context."""
    try:
        return __state.get()
    except LookupError as e:
        raise InvalidContext() from e


def current_graph() -> TierkreisGraph:
    """The current GraphBuilder's graph."""
    return current_builder().graph


def _set_state(bg: "GraphBuilder") -> Token["GraphBuilder"]:
    return __state.set(bg)


def _reset_state(tok: Token["GraphBuilder"]):
    return __state.reset(tok)


def _capture_label(idx: int) -> PortID:
    return f"_c{idx}"


@dataclass(frozen=True)
class _CaptureOutwards:
    # Utility to allow capturing values from inner scopes to outer scopes.

    source_graph: TierkreisGraph
    ref: NodeRef
    contained: Optional["_CaptureOutwards"] = None

    def capture(self, incoming: ValueSource) -> ValueSource:
        """Convert a ValueSource from the inner scope to the outer scope by adding an output to the inner."""

        if self.contained is not None:
            incoming = self.contained.capture(incoming)
        new_name = _capture_label(len(self.source_graph.outputs()))
        self.source_graph.set_outputs(**{new_name: _resolve(incoming)})
        return self.ref[new_name]


TGBuilder = TypeVar("TGBuilder", bound="GraphBuilder")


class GraphBuilder(AbstractContextManager):
    """Builder context that maintains a graph state and allows adding nodes
    within the context."""

    graph: TierkreisGraph
    inputs: list[str]
    outputs: list[str]
    inner_scopes: dict[TierkreisGraph, _CaptureOutwards]
    debug_info: dict[NodeRef, str]
    _state_token: Token["GraphBuilder"]

    def __init__(
        self,
        inputs: list[str] | None = None,
        outputs: list[str] | None = None,
        name: str = "",
        debug: bool = True,
    ) -> None:
        """Create a new GraphBuilder context.

        Args:
            inputs: Optional list of input port names, defaults to None (i.e. the empty list).
            outputs: Optional list of output port names, defaults to None (i.e. the empty list).
            name: _description_. Name the graph being built.
            debug: _description_. Whether to store debug info while building..
        """
        self.graph = TierkreisGraph(name)
        self.inputs = inputs or []
        self.outputs = outputs or []
        self.inner_scopes = {}
        self.graph.input_order.extend(self.inputs)
        self.debug_info = {self.graph.input: _debug_str()} if debug else {}

    def input(self, name: str) -> NodePort:
        """Get a port from the input node by name."""
        if name in self.inputs:
            return self.graph.input[name]
        raise RuntimeError("Input not declared.")

    def __enter__(self: TGBuilder) -> TGBuilder:
        self._state_token = _set_state(self)
        return self

    def __exit__(
        self,
        __exc_type: Optional[Type[BaseException]],
        __exc_value: Optional[BaseException],
        __traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        _reset_state(self._state_token)
        return None

    def _inner_capture(self, incoming: ValueSource) -> Optional[ValueSource]:
        source_graph = _source_graph(incoming)
        if source_graph is self.graph:
            return incoming
        inner_scope = self.inner_scopes.get(source_graph)
        if inner_scope is not None:
            return inner_scope.capture(incoming)
        return None

    def capture(
        self, incoming: ValueSource, allow_existing: bool = False
    ) -> ValueSource:
        """Capture an input from an outer scope."""
        if (ret_val := self._inner_capture(incoming)) is None:
            raise InvalidContext(
                f"ValueSource {incoming} does not belong to current graph."
            )
        return ret_val

    def _add_edges(
        self,
        nr: NodeRef,
        /,
        **incoming_wires: ValueSource,
    ) -> None:
        for tgt_port, vs in incoming_wires.items():
            src = _resolve(self.capture(vs))
            self.graph.add_edge(src, nr[tgt_port])

    def add_node_to_graph(
        self,
        _tk_node: TierkreisNode,
        /,
        **incoming_wires: ValueSource,
    ) -> NodeRef:
        """Add a node to the graph and connect incoming wires, which are
        specified via keyword arguments."""
        nr = self.graph.add_node(_tk_node)
        self._add_edges(nr, **incoming_wires)
        if self.debug_info:
            self.debug_info[nr] = _debug_str()
        return nr

    def set_graph_outputs(self, **incoming_wires: ValueSource) -> None:
        """Set the outputs of the graph by specifying incoming wires as keyword arguments."""
        if len(self.graph.outputs()) != 0:
            raise RuntimeError("Outputs set multiple times for same graph.")
        self._add_edges(self.graph.output, **incoming_wires)

    def type_check(self, sig: Signature) -> TierkreisGraph:
        """Type check a graph against a provided signature and return a new graph with
        type annotations. Throws an exception if  `typecheck` optional
        dependencies not installed)."""
        try:
            return infer_graph_types(self.graph, sig)
        except TierkreisTypeErrors as te:
            raise te.with_debug(self.debug_info) from None


class CaptureBuilder(GraphBuilder):
    """Graph builder for closures which capture inputs from outer scopes."""

    captured: dict[ValueSource, PortID]

    def __init__(
        self,
        inputs: list[str] | None = None,
        outputs: list[str] | None = None,
        name: str = "",
        debug: Optional[bool] = None,
    ) -> None:
        bool_debug = bool(current_builder().debug_info) if debug is None else debug
        super().__init__(inputs, outputs, name, bool_debug)

        self.captured = {}

    def capture(
        self, incoming: ValueSource, allow_existing: bool = False
    ) -> ValueSource:
        """Capture input from an outer scope, wiring it through the input
        node.
        """
        if (valsrc := self._inner_capture(incoming)) is not None:
            return valsrc
        if incoming in self.captured and not allow_existing:
            # This would generate two wires from the same NodePort
            raise ValueError(f"Already captured {incoming} as input to {self.graph}")
        in_name = self.captured.setdefault(incoming, _capture_label(len(self.captured)))
        return self.graph.input[in_name]

    def __exit__(
        self,
        __exc_type: Optional[Type[BaseException]],
        __exc_value: Optional[BaseException],
        __traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        super().__exit__(__exc_type, __exc_value, __traceback)
        if not any((__exc_type, __exc_value, __traceback)):
            current_builder().debug_info.update(self.debug_info)
        return None


def _partial_thunk(
    thunk: Union[GraphBuilder, NodePort], captured: dict[ValueSource, PortID]
) -> NodePort:
    # add the thunk as constant to the current graph and partial up
    bg = current_builder()
    thunk_port: NodePort = (
        Const(thunk.graph) if isinstance(thunk, GraphBuilder) else thunk
    )
    if captured:
        return bg.add_node_to_graph(
            FunctionNode(FunctionName("partial")),
            thunk=thunk_port,
            **{k: v for v, k in captured.items()},
        )[Labels.VALUE]
    return thunk_port


T = TypeVar("T")


def Const(
    val: Any,
    type_: Type | None = None,
) -> NodePort:
    """Add a constant value to the graph. If a type is provided, it will be used
    in the conversion, otherwise the type will be inferred from the value."""
    if isinstance(val, LazyGraph):
        val = val.graph
    n = current_builder().add_node_to_graph(
        ConstNode(TierkreisValue.from_python(val, type_))
    )
    return n[Labels.VALUE]


def UnionConst(val: Any) -> NodePort:
    """Add a constant as a variant value tagged by its type, to be used
    with `Union` type annotations.
    """
    union_tag = UnionTag.value_type_tag(val)
    return Tag(union_tag, Const(val))


@dataclass(frozen=True)
class _CallAddNode:
    node: TierkreisNode
    input_order: list[str]
    output_order: list[str]

    def __call__(self, *args: ValueSource, **kwds: ValueSource) -> NodeRef:
        bg = current_builder()
        kwds = _combine_args_with_kwargs(self.input_order, *args, **kwds)

        n = bg.add_node_to_graph(self.node, **kwds)
        return NodeRef(n.idx, n.graph, self.output_order)


class Scope(CaptureBuilder, ABC):
    """Build a sub graph to be executed at a specific `Location`."""

    location: Location

    def __init__(self, location: Union[str, Location] = Location([])):
        if isinstance(location, str):
            self.location = Location(location.split("/"))
        else:
            self.location = location
        super().__init__()

    def __exit__(
        self,
        __exc_type: Optional[Type[BaseException]],
        __exc_value: Optional[BaseException],
        __traceback: Optional[TracebackType],
    ) -> None:
        super().__exit__(__exc_type, __exc_value, __traceback)

        if not any((__exc_type, __exc_value, __traceback)):
            graph = self.graph
            inputs = {v: k for k, v in self.captured.items()}
            node_ref = _build_box(graph, self.location, **inputs)
            for inner_graph, captured in self.inner_scopes.items():
                current_builder().inner_scopes[inner_graph] = _CaptureOutwards(
                    graph, node_ref, captured
                )
            current_builder().inner_scopes[graph] = _CaptureOutwards(graph, node_ref)


def Tag(tag: str, value: ValueSource) -> NodePort:
    """Add a tag node to the graph, tagging the value with the provided tag
    string and producing a :class:`~tierkreis.core.values.VariantValue` from the output port.
    """
    return current_builder().add_node_to_graph(TagNode(tag), value=value)[Labels.VALUE]


class Output(NodeRef):
    """Set the outputs of a graph to the provided values. Return this object
    when terminating a graph building context. Positional arguments are mapped to any output
    names annotated on the graph being built, keyword arguments are used to
    explicitly set outputs.
    """

    def __init__(self, *args: ValueSource, **kwargs: ValueSource) -> None:
        s = current_builder()

        g = s.graph

        if len(args) == 1 and len(kwargs) == 0:
            kwargs = {Labels.VALUE: args[0]}
        else:
            kwargs = _combine_args_with_kwargs(s.outputs, *args, **kwargs)

        s.set_graph_outputs(**kwargs)
        for o in kwargs:
            g.output_order.append(o)

        super().__init__(g.output.idx, g)


class _LoopOutput(Output):
    def __init__(self, loopout: ValueSource) -> None:
        super().__init__(value=loopout)


class Continue(_LoopOutput):
    """Exit the current iteration of a loop and continue to the next with the
    provided output.
    """

    def __init__(self, loopout: ValueSource) -> None:
        super().__init__(Tag(Labels.CONTINUE, loopout))


class Break(_LoopOutput):
    """Exit the loop with the provided output."""

    def __init__(self, loopout: ValueSource) -> None:
        super().__init__(Tag(Labels.BREAK, loopout))


def _input_port(port: PortID) -> NodePort:
    return current_builder().graph.input[port]


_GraphDef = Callable[..., Output]


def _get_input_names(f: _GraphDef) -> list[str]:
    return list(inspect.signature(f).parameters.keys())


@dataclass(frozen=True)
class LazyGraph:
    """Context to lazily build a graph from a decorated function. A graph is only
    built when the function is called.
    """

    _builder: Callable[[], TierkreisGraph]

    @cached_property
    def graph(self) -> TierkreisGraph:
        return self._builder()

    def __call__(self, *args: ValueSource, **kwargs: ValueSource) -> NodeRef:
        return self.graph(*args, **kwargs)

    def _repr_svg_(self) -> str:
        return getattr(self.graph, "_repr_svg_")()


_GraphDecoratorType = Callable[[_GraphDef], TierkreisGraph]
_LazyGraphDecoratorType = Callable[[_GraphDef], LazyGraph]


def lazy_graph(
    name: str | None = None,
    type_check_sig: Signature | None = None,
    output_order: list[str] | None = None,
    debug: bool = True,
) -> _LazyGraphDecoratorType:
    """Decorate a function to build a `TierkreisGraph` when it is called.

    Args:
        name: Optionally name the graph,
            defaults to None
        type_check_sig: If a type signature
            is provided, the graph will be type checked when built,
            defaults to None
        output_order: Optional list of
            output ports, inputs are inferred from function arguments,
            defaults to None
    """

    def decorator_graph(f: _GraphDef) -> LazyGraph:
        input_order = _get_input_names(f)
        # Use getattr as _GraphDef doesn't specify that it has a __name__
        graph_name = name or getattr(f, "__name__")
        # graph_name = name or f.__name__  # type: ignore  # the alternative

        def wrapper() -> TierkreisGraph:
            gb = GraphBuilder(input_order, output_order, name=graph_name, debug=debug)

            with gb as sub_build:
                f(*(_input_port(port) for port in input_order))

            return (
                sub_build.graph
                if type_check_sig is None
                else sub_build.type_check(type_check_sig)
            )

        return LazyGraph(wrapper)

    return decorator_graph


def graph(
    name: Optional[str] = None,
    type_check_sig: Optional[Signature] = None,
    output_order: list[str] | None = None,
    debug: bool = True,
) -> _GraphDecoratorType:
    """Convert a function into a `TierkreisGraph`. See `lazy_graph` for details.
    This function is equivalent except that the graph is built when the module is loaded and replaces
    the original function.
    """
    dec = lazy_graph(name, type_check_sig, output_order, debug)

    def decorator_graph(f: _GraphDef) -> TierkreisGraph:
        return dec(f).graph

    return decorator_graph


class Thunk(_CallAddNode, PortFunc):
    """Utility to call a thunk (a graph) as a function. Adds an eval node when called."""

    graph_src: ValueSource

    def __init__(
        self, graph_src: ValueSource, input_order: list[str], output_order: list[str]
    ):
        self.graph_src = graph_src
        super().__init__(FunctionNode(FunctionName("eval")), input_order, output_order)

    def __call__(self, *args: ValueSource, **kwargs: ValueSource) -> NodeRef:
        assert "thunk" not in kwargs
        return super().__call__(*args, thunk=self.graph_src, **kwargs)

    def copyable(self) -> "Thunk":
        self.graph_src = Copyable(self.graph_src)
        return self

    def resolve(self) -> NodePort:
        """Return a NodePort, creating nodes as necessary in source graph."""
        return _resolve(self.graph_src)

    def source_graph(self) -> TierkreisGraph:
        """The graph the port belongs to (or will)."""
        return _source_graph(self.graph_src)


def closure(
    name: Optional[str] = None,
    debug: bool = True,
    output_order: list[str] | None = None,
) -> Callable[[_GraphDef], Thunk]:
    """Decorator to build a closure `TierkreisGraph` inside a builder context.

    Args:
        name: Optionally name the closure,
            defaults to None
        output_order: Optional list of
            output ports, defaults to None
    """

    def decorator_graph(f: _GraphDef) -> Thunk:
        input_order = _get_input_names(f)

        # Use getattr as _GraphDef doesn't specify that it has a __name__
        graph_name = name or getattr(f, "__name__")
        # graph_name = name or f.__name__  # type: ignore  # the alternative

        gb = CaptureBuilder(input_order, output_order, name=graph_name, debug=debug)

        with gb as sub_build:
            f(*(_input_port(port) for port in input_order))
        return Thunk(_partial_thunk(sub_build, sub_build.captured), input_order, [])

    return decorator_graph


def loop(name: Optional[str] = None, debug: bool = True):
    """Decorator to build a loop body `TierkreisGraph` inside a builder context.
    Must have exactly one input argument, the loop variable.
    """

    def decorator_graph(f: _GraphDef):
        inps = _get_input_names(f)
        if len(inps) != 1:
            raise ValueError("Loop body graph can only have one input.")

        # the loop node has input port "value"
        # allow body definitions to use any input names and map it
        graph_name = name or getattr(f, "__name__")

        def wrapper(initial: ValueSource) -> NodePort:
            gb = CaptureBuilder(inps, None, name=graph_name, debug=debug)
            with gb as sub_build:
                f(_input_port(Labels.VALUE))

            loop_node = current_builder().add_node_to_graph(
                FunctionNode(FunctionName("loop")),
                body=_partial_thunk(sub_build, sub_build.captured),
                value=initial,
            )
            return loop_node[Labels.VALUE]

        return wrapper

    return decorator_graph


class If(CaptureBuilder):
    """Utility to build an if/else block, adds a `Match` node when called."""

    def __init__(self) -> None:
        super().__init__()
        self.graph.name = "if"
        Match._add_handler("if", self)


class Else(CaptureBuilder):
    """Utility to build an if/else block, adds a `Match` node when called."""

    def __init__(self) -> None:
        super().__init__()
        self.graph.name = "else"
        Match._add_handler("else", self)


class DuplicateBlock(Exception):
    """Duplicate blocks, avoid multiple If/Else or duplicate Case blocks."""

    block_name: str


class IntermediateNodes(Exception):
    """Adding of nodes not valid in _CaseScope outside child blocks."""


class _CaseScope(GraphBuilder, ABC):
    __variant_handlers: ContextVar[dict[str, CaptureBuilder]] = ContextVar(
        "variant_handlers"
    )

    _token: Token[dict[str, CaptureBuilder]]
    nref: NodeRef

    @classmethod
    def _get_handlers(cls) -> dict[str, CaptureBuilder]:
        return cls.__variant_handlers.get()

    def __enter__(self) -> "_CaseScope":
        super().__enter__()
        self._token = self.__variant_handlers.set({})
        return self

    @classmethod
    def _add_handler(cls, tag: str, handler: CaptureBuilder):
        handlers = cls.__variant_handlers.get()
        if tag in handlers:
            raise DuplicateBlock(tag)
        handlers[tag] = handler

    @abstractmethod
    def _get_thunk(self) -> NodePort: ...

    @abstractmethod
    def _intermediate_error(self) -> IntermediateNodes: ...

    def __exit__(
        self,
        __exc_type: Optional[Type[BaseException]],
        __exc_value: Optional[BaseException],
        __traceback: Optional[TracebackType],
    ) -> None:
        super().__exit__(__exc_type, __exc_value, __traceback)

        if not any((__exc_type, __exc_value, __traceback)):
            # do not perform graph building if error

            if self.graph.n_nodes != 2:
                raise self._intermediate_error()
            handlers = self.__variant_handlers.get()
            for h in handlers.values():
                current_builder().debug_info.update(h.debug_info)

            thunk_port = self._get_thunk()
            partial_inps = _combine_captures(list(handlers.values()))
            thunk_port = _partial_thunk(thunk_port, partial_inps)

            self.nref = current_builder().add_node_to_graph(
                FunctionNode(FunctionName("eval")),
                thunk=thunk_port,
            )
        self.__variant_handlers.reset(self._token)


class Match(_CaseScope):
    """Pattern match on a variant value by providing a `Case` for each variant."""

    variant_value: ValueSource

    def __init__(self, variant_value: ValueSource):
        super().__init__()
        self.variant_value = variant_value

    def _get_thunk(self) -> NodePort:
        handlers = _CaseScope._get_handlers()
        bg = current_builder()
        return bg.add_node_to_graph(
            MatchNode(),
            variant_value=self.variant_value,
            **map_vals(handlers, lambda x: Const(x.graph)),
        )[Labels.THUNK]

    def _intermediate_error(self) -> IntermediateNodes:
        return IntermediateNodes(
            "Graph building operations outside Case blocks not allowed in Match block."
        )


class Case(CaptureBuilder):
    """Define a graph to handle a specific variant in a `Match` block."""

    var_value: NodePort

    def __init__(self, tag: str) -> None:
        super().__init__()
        self.graph.name = tag
        Match._add_handler(tag, self)

        self.var_value = self.graph.input[Labels.VALUE]


class IfElse(_CaseScope):
    """Define an if/else block. Must contain exactly one `If` block and exactly one `Else` blocks."""

    predicate: ValueSource

    def __init__(self, predicate: ValueSource):
        super().__init__()
        self.predicate = predicate

    def _get_thunk(self) -> NodePort:
        handlers = _CaseScope._get_handlers()
        assert len(handlers) == 2
        bg = current_builder()
        return bg.add_node_to_graph(
            FunctionNode(FunctionName("switch")),
            pred=self.predicate,
            if_true=Const(handlers["if"].graph),
            if_false=Const(handlers["else"].graph),
        )[Labels.VALUE]

    def _intermediate_error(self) -> IntermediateNodes:
        return IntermediateNodes(
            "Graph building operations outside"
            " If/Else blocks not allowed in IfElse block."
        )


def _combine_captures(thunks: list[CaptureBuilder]) -> dict[ValueSource, PortID]:
    """Given a list of thunks with captured inputs, remove their existing input
    names, common up any shared sources (up to == equality), and add all those
    inputs to all graphs, discarding any unused.
    """
    old_names = frozenset([v for bg in thunks for v in bg.captured.values()])
    # Generate a sparsely numbered list of names that won't conflict:
    names = filter(lambda x: x not in old_names, map(_capture_label, count()))
    # Assign new names in deterministic order of thunks and captures in each
    combined_inputs: dict[ValueSource, str] = {
        vs: next(names) for bg in thunks for vs in bg.captured
    }  # Duplicate ValueSources overwrite earlier entries, making even sparser.
    for vs, nam in combined_inputs.items():
        for bg in thunks:
            if in_name := bg.captured.get(vs):
                in_edge = bg.graph.out_edge_from_port(bg.graph.input[in_name])
                assert in_edge is not None
                bg.graph._graph.remove_edge(*in_edge.to_edge_handle())
                bg.graph.add_edge(bg.graph.input[nam], in_edge.target)
            else:
                bg.graph.discard(bg.graph.input[nam])

    return combined_inputs


def _combine_args_with_kwargs(
    expected: Iterable[str], *args: ValueSource, **kwargs: ValueSource
) -> dict[str, ValueSource]:
    pos_args = dict(zip(expected, args))
    if ints := set(kwargs).intersection(pos_args):
        raise RuntimeError(f"Arguments specified positionally and in keywords: {ints}")
    kwargs.update(pos_args)
    return kwargs


class Copyable(PortFunc):
    """Inserts a copy of an underlying ValueSource whenever an outgoing edge is added
    unless there are *no* existing uses of the underlying ValueSource
    (=> the wire is routed from the underlying ValueSource)
    """

    np: NodePort

    def __init__(self, src: ValueSource):
        builder = current_builder()
        while isinstance(src, Copyable) and _source_graph(src) == builder.graph:
            # Simplify if we call Copyable on a ValueSource twice
            self = src
        self.np = _resolve(builder.capture(src, allow_existing=True))

    def resolve(self) -> NodePort:
        g = self.source_graph()
        existing_edge = g.out_edge_from_port(self.np)
        if existing_edge is None:
            return self.np
        g._graph.remove_edge(*existing_edge.to_edge_handle())
        c1, c2 = g.copy_value(self.np)
        g.add_edge(c1, existing_edge.target, existing_edge.type_)
        return c2

    def source_graph(self) -> "TierkreisGraph":
        return self.np.node_ref.graph


@dataclass(frozen=True)
class Unpack(StablePortFunc):
    """Utility class representing an unpacked field of a struct."""

    src: Union[NodePort, StablePortFunc]
    field_name: str

    def resolve(self) -> NodePort:
        np = _resolve(self.src)
        g = np.node_ref.graph
        existing_edge = g.out_edge_from_port(np)
        if existing_edge is not None:
            target_node = g[existing_edge.target.node_ref]
            if not target_node.is_unpack_node():
                raise ValueError(
                    "Cannot unpack port wired to something "
                    f" other than an unpack_struct: {target_node}"
                )
            unpack_node = existing_edge.target.node_ref
        else:
            unpack_node = g.add_func("unpack_struct", struct=np)
        return unpack_node[self.field_name]

    def source_graph(self) -> "TierkreisGraph":
        return _source_graph(self.src)


def __unpack_nodeport(self: NodePort, field_name: str) -> Unpack:
    return Unpack(self, field_name)


NodePort.__getitem__ = __unpack_nodeport  # type: ignore


def _arg_str(args: Dict[str, "TierkreisType"], order: Iterable[str]) -> str:
    return ", ".join(f"{chalk.yellow(port)}: {args[port]}" for port in order)


def _func_sig(name: str, func: FunctionDeclaration):
    graph_type = cast("GraphType", TypeScheme.from_proto(func.type_scheme).body)
    irest = graph_type.inputs.rest
    orest = graph_type.outputs.rest
    irest = f", {chalk.yellow('#')}: {irest}" if irest else ""
    orest = f", {chalk.yellow('#')}: {orest}" if orest else ""
    return (
        f"{chalk.bold.blue(name)}"
        f"({_arg_str(graph_type.inputs.content, func.input_order)}{irest})"
        f" -> ({_arg_str(graph_type.outputs.content, func.output_order)}{orest})"
    )


class Function(_CallAddNode):
    """A Tierkreis function - i.e. provided by a worker, or a builtin."""

    f: FunctionDeclaration
    name: FunctionName

    def __init__(self, f: FunctionDeclaration, name: FunctionName):
        self.f = f
        self.name = name
        super().__init__(FunctionNode(name), f.input_order, f.output_order)

    def signature_string(self) -> str:
        return _func_sig(self.name.name, self.f)

    def to_graph(
        self,
        input_map: Optional[dict[str, str]] = None,
        output_map: Optional[dict[str, str]] = None,
    ) -> TierkreisGraph:
        return graph_from_func(self.name, self.f, input_map, output_map)

    def __str__(self) -> str:
        return self.signature_string()


class Namespace(Mapping[str, "Namespace"]):
    """Namespace of functions and sub-namespaces."""

    @overload
    def __init__(self, args: Signature, /): ...

    @overload
    def __init__(self, args: SigNamespace, *, _prefix: list[str]):
        """For internal use only."""
        ...

    def __init__(self, args: Union[Signature, SigNamespace], **kwargs: list[str]):
        if isinstance(args, Signature):
            namespace = args.root
            prefix = []
        else:
            namespace = args
            prefix = kwargs["_prefix"]
        for name, ndef in namespace.functions.items():
            # allows auto completion of function names in Jupyter environment
            setattr(self, name, Function(ndef, FunctionName(name, prefix)))
        self.__subspaces: Mapping[str, "Namespace"] = {
            name: Namespace(ns, _prefix=prefix + [name])
            for name, ns in namespace.subspaces.items()
        }

    def __getattr__(
        self, attr_name: str
    ) -> Function:  # This is here just to placate mypy
        raise AttributeError(attr_name)

    def __getitem__(self, __k: str) -> "Namespace":
        return self.__subspaces[__k]

    def __len__(self) -> int:
        return self.__subspaces.__len__()

    def __iter__(self) -> Iterator[str]:
        return self.__subspaces.__iter__()


def _rename_ports(
    thunk: ValueSource, map_old_to_new: dict[str, str], is_inputs: bool
) -> NodePort:
    bg = current_builder()
    rename_g = Const(rename_ports_graph(map_old_to_new))
    f, s = (rename_g, thunk) if is_inputs else (thunk, rename_g)
    seq = bg.add_node_to_graph(
        FunctionNode(FunctionName("sequence")), first=f, second=s
    )

    return seq["sequenced"]


def RenameInputs(thunk: ValueSource, rename_map: dict[str, str]) -> NodePort:
    """Utility function to rename the inputs of a thunk (runtime graph value)."""
    return _rename_ports(thunk, {v: k for k, v in rename_map.items()}, True)


def RenameOutputs(thunk: ValueSource, rename_map: dict[str, str]) -> NodePort:
    """Utility to rename the outputs of a thunk (runtime graph value)."""
    return _rename_ports(thunk, rename_map, False)


def _build_box(
    g: TierkreisGraph, __tk_loc: Location, *args: ValueSource, **kwargs: ValueSource
) -> NodeRef:
    bx = _CallAddNode(
        BoxNode(g, __tk_loc),
        g.input_order,
        g.output_order,
    )
    return bx(*args, **kwargs)


def __graph_call__(
    self: TierkreisGraph, *args: ValueSource, **kwargs: ValueSource
) -> NodeRef:
    return _build_box(self, Location([]), *args, **kwargs)


def UnpackTuple(src: ValueSource, first_n: int) -> Iterator[NodePort]:
    """Unpack a struct containing python tuple fields, up to the `first_n` fields."""
    unpack_node = current_builder().add_node_to_graph(
        FunctionNode(FunctionName("unpack_struct")),
        struct=src,
    )

    return (unpack_node[TupleLabel.index_label(i)] for i in range(first_n))


def MakeTuple(*args: ValueSource) -> NodePort:
    """Make a struct from some values, treating them like a python tuple, with
    member position indices used to generate the field labels.
    """
    make_node = current_builder().add_node_to_graph(
        FunctionNode(FunctionName("make_struct")),
        **{TupleLabel.index_label(i): v for i, v in enumerate(args)},
    )
    return make_node["struct"]


# mypy is not happy with assignment to a method
setattr(TierkreisGraph, "__call__", __graph_call__)



"""Implementation of `tkrs` CLI tool, for command line graph execution and visualisation."""

import asyncio
import re
import sys
import traceback
from contextlib import asynccontextmanager
from functools import wraps
from pathlib import Path
from signal import SIGINT, SIGTERM
from typing import TYPE_CHECKING, AsyncContextManager, Dict, List, Optional, cast

import click
from grpclib.client import Channel
from yachalk import chalk

import tierkreis.core.protos.tierkreis.v1alpha1.graph as pg
from tierkreis import TierkreisGraph
from tierkreis.builder import _func_sig
from tierkreis.client import ServerRuntime
from tierkreis.core.graphviz import tierkreis_to_graphviz
from tierkreis.core.protos.tierkreis.v1alpha1.graph import Graph as ProtoGraph
from tierkreis.core.signature import Signature
from tierkreis.core.type_errors import TierkreisTypeErrors
from tierkreis.core.types import StructType
from tierkreis.core.values import StructValue, TierkreisValue

if TYPE_CHECKING:
    from tierkreis.core.function import FunctionDeclaration


@asynccontextmanager
async def server_manager(host: str, port: int = 443):
    async with Channel(host, port) as channel:
        yield ServerRuntime(channel)


def coro(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        return asyncio.run(f(*args, **kwargs))

    return wrapper


def _inputs(proto_file: str, py_source: str) -> Dict[str, TierkreisValue]:
    if proto_file == "":
        return (
            {}
            if py_source == ""
            else {k: TierkreisValue.from_python(v) for k, v in eval(py_source).items()}
        )
    if py_source != "":
        raise ValueError("Cannot provide both inputs and --py-inputs")
    path = Path(proto_file)
    with open(path, "rb") as f:
        v: StructValue = StructValue.from_proto(pg.StructValue().parse(f.read()))
        return v.values


async def _parse(source: Path) -> TierkreisGraph:
    with open(source, "rb") as f:
        return TierkreisGraph.from_proto(ProtoGraph().parse(f.read()))


async def _check_graph(
    source_path: Path,
    client_manager: AsyncContextManager[ServerRuntime],
) -> TierkreisGraph:
    async with client_manager as client:
        tkg = await _parse(source_path)
        try:
            tkg = await client.type_check_graph(tkg)
        except TierkreisTypeErrors:
            _print_typeerrs(traceback.format_exc(0))
            sys.exit(1)
        return tkg


async def main_coro(manager: AsyncContextManager):
    async with manager as runtime:
        try:
            print(runtime.socket_address(), flush=True)
            await asyncio.sleep(10000000000000)
        except asyncio.CancelledError:
            print("\nShutting Down")


def run_with_signals(manager: AsyncContextManager):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    main_task = loop.create_task(main_coro(manager))
    for signal in [SIGINT, SIGTERM]:
        loop.add_signal_handler(signal, main_task.cancel)
    try:
        loop.run_until_complete(main_task)
    finally:
        loop.close()


@click.group()
@click.pass_context
@click.option(
    "--runtime",
    "-r",
    default="localhost",
    help="Choose runtime, default=localhost",
)
@click.option(
    "--port",
    "-p",
    help="Runtime port, default=8090 if runtime is localhost, else 443",
)
@coro
async def cli(ctx: click.Context, runtime: str, port: Optional[int]):
    local = runtime == "localhost"
    if port is None:
        port = 8090 if local else 443
    ctx.ensure_object(dict)
    ctx.obj["runtime_label"] = runtime
    client_manager = server_manager(runtime, port)
    asyncio.get_event_loop()
    ctx.obj["client_manager"] = client_manager


@cli.command()
@click.argument("proto", type=click.Path(exists=True))
@click.pass_context
@coro
async def check(ctx: click.Context, proto: str) -> TierkreisGraph:
    """Type check PROTO binary file against runtime signature."""
    source_path = Path(proto)
    tkg = await _check_graph(source_path, ctx.obj["client_manager"])
    print(chalk.bold.green("Success: graph type check complete."))
    return tkg


@cli.command()
@click.argument("proto", type=click.Path(exists=True))
@click.argument(
    "view_path",
    type=click.Path(exists=False),
)
@click.option("--check", "-C", is_flag=True, help="Type check and annotate graph.")
@click.option(
    "--unbox-level",
    default=0,
    help="Nesting level to which boxes and thunks should be unthunked, default=0.",
)
@click.pass_context
@coro
async def view(
    ctx: click.Context,
    proto: str,
    view_path: str,
    check: bool,
    unbox_level: int,
):
    """Visualise PROTO binary as graph and output to VIEW_PATH."""
    source_path = Path(proto)
    if check:
        tkg = await _check_graph(
            source_path,
            ctx.obj["client_manager"],
        )
    else:
        tkg = await _parse(source_path)

    tkg.name = source_path.stem
    view_p = Path(view_path)
    ext = view_p.suffix
    tierkreis_to_graphviz(tkg, unbox_level=unbox_level).render(
        view_path[: -len(ext)], format=ext[1:]
    )


def _print_outputs(outputs: dict[str, TierkreisValue]):
    print(
        "\n".join(
            f"{chalk.bold.yellow(key)}: {val.viz_str()}" for key, val in outputs.items()
        )
    )


def _print_typeerrs(errs: str):
    print(chalk.red(errs), file=sys.stderr)


@cli.command()
@click.argument("proto", type=click.Path(exists=True))
@click.argument("inputs", default="")
@click.option("-p", "--py-inputs", default="")
@click.pass_context
@coro
async def run(ctx: click.Context, proto: Path, inputs: str, py_inputs: str):
    """Run PROTO binary on runtime with optional INPUTS and output to console."""
    async with ctx.obj["client_manager"] as client:
        client = cast(ServerRuntime, client)
        tkg = await _parse(proto)
        input_dict = _inputs(inputs, py_inputs)
        try:
            outputs = await client.run_graph(tkg, **input_dict)
            _print_outputs(outputs)
        except TierkreisTypeErrors:
            _print_typeerrs(traceback.format_exc(0))
            sys.exit(1)


PORT_RE = re.compile(r"([\w]+):")


def _print_namespace(sig: Signature, namespace: List[str], function: Optional[str]):
    namespace_str = "::".join(namespace)
    root = "_root"

    print(chalk.bold(f"Namespace: {root if namespace == [] else namespace_str}"))
    print()
    print(chalk.bold("Aliases and Struct definitions"))

    for alias, type_scheme in sig.aliases.items():
        prefix = f"{namespace_str}::"
        if not alias.startswith(prefix):
            continue
        alias = alias[len(prefix) :]
        type_ = type_scheme.body
        if isinstance(type_, StructType):
            alias_string = type_.anon_name()
            alias_string = PORT_RE.sub(
                lambda match: chalk.yellow(match.group()), alias_string
            )
        else:
            alias_string = str(type_)
        print(f"{chalk.bold.magenta(alias)} = {alias_string}\n")

    print()
    print(chalk.bold("Functions"))

    names_dict = sig.root.get(namespace).functions
    func_names = [function] if function else list(names_dict.keys())
    for name in sorted(func_names):
        _print_func(name, names_dict[name])
        print()


def _print_func(name: str, func: "FunctionDeclaration"):
    print(_func_sig(name, func))
    if func.description:
        print(chalk.green(func.description))


@cli.command()
@click.pass_context
@click.option("--namespace", type=str, help="Show only signatures of this namespace.")
@click.option(
    "--function", type=str, help="Show only the signature of a particular function."
)
@coro
async def signature(
    ctx: click.Context, namespace: Optional[str], function: Optional[str]
):
    """Check signature of available namespaces and functions on runtime."""
    async with ctx.obj["client_manager"] as client:
        client = cast(ServerRuntime, client)
        label = ctx.obj["runtime_label"]
        print(chalk.bold(f"Runtime: {label}"))
        print()
        sig = await client.get_signature()
        namespaces: list[list[str]] = (
            [namespace.split("::")] if namespace else sig.root.all_namespaces()
        )

        for ns in namespaces:
            _print_namespace(sig, ns, function)
            print()



import platform
import sys
from copy import deepcopy
from typing import TYPE_CHECKING, Iterator

import betterproto
import pytest

from tierkreis import TierkreisGraph
from tierkreis.builder import (
    Break,
    Const,
    Continue,
    Else,
    If,
    IfElse,
    Output,
    Scope,
    ValueSource,
    closure,
    graph,
    lazy_graph,
    loop,
)
from tierkreis.client import RuntimeClient, ServerRuntime
from tierkreis.core.function import FunctionDeclaration
from tierkreis.core.signature import Namespace, Signature
from tierkreis.core.tierkreis_graph import GraphValue, NodePort
from tierkreis.core.type_errors import TierkreisTypeErrors, UnifyError, UnknownFunction
from tierkreis.core.type_inference import infer_graph_types
from tierkreis.core.types import (
    FloatType,
    GraphType,
    IntType,
    PairType,
    Row,
    StarKind,
    TypeScheme,
    VarType,
)
from tierkreis.core.values import FloatValue, StructValue

if TYPE_CHECKING:
    from tierkreis.builder import Namespace as BuilderNS


@pytest.mark.asyncio
async def test_infer(client: RuntimeClient) -> None:
    # test when built with client types are auto inferred
    tg = TierkreisGraph()
    _, val1 = tg.copy_value(tg.add_const(3))
    tg.set_outputs(out=val1)
    if not client.can_type_check:
        pytest.skip()
    tg = await client.type_check_graph(tg)
    assert any(node.is_discard_node() for node in tg.nodes())

    assert isinstance(tg.get_edge(val1, NodePort(tg.output, "out")).type_, IntType)


@pytest.mark.asyncio
async def test_infer_errors(client: RuntimeClient) -> None:
    # build graph with two type errors
    tg = TierkreisGraph()
    node_0 = tg.add_const(0)
    node_1 = tg.add_const(1)
    tg.add_edge(node_0["value"], tg.input["illegal"])
    tg.set_outputs(out=node_1["invalid"])

    if not client.can_type_check:
        pytest.skip()

    with pytest.raises(
        TierkreisTypeErrors, match="There are extra inputs: {'illegal'}"
    ) as err1:
        await client.type_check_graph(tg)
    returned_errs = [err1.value.errors]
    if isinstance(client, ServerRuntime):
        # only server runtime type checks on run
        with pytest.raises(TierkreisTypeErrors) as err:
            await client.run_graph(tg)
        returned_errs.append(err.value.errors)
    for errs in returned_errs:
        assert len(errs) == 2
        assert all(isinstance(e, UnifyError) for e in errs)

        # TODO make type errors report in deterministic order
        locs = (e.proto_err.location for e in errs)

        locations = {
            betterproto.which_one_of(loc[1], "location")[0]: loc[1] for loc in locs
        }
        assert set(locations.keys()) == {"input", "node_idx"}
        assert locations["node_idx"].node_idx == node_1.idx


_foo_func = FunctionDeclaration(
    TypeScheme(
        {"a": StarKind()},
        [],
        GraphType(
            inputs=Row({"value": VarType("a")}, None),
            outputs=Row({"res": PairType(VarType("a"), IntType())}, None),
        ),
    ).to_proto(),
    "no docs",
    ["value"],
    ["res"],
)


@pytest.mark.skip_typecheck
def test_infer_graph_types():
    tg = TierkreisGraph()
    foo = tg.add_func("foo", value=tg.add_const(3))
    tg.set_outputs(out=foo["res"])
    with pytest.raises(TierkreisTypeErrors, match="Unknown function: foo") as err:
        infer_graph_types(tg, Signature.empty())
    assert len(err.value) == 1
    e = err.value.errors[0]
    assert isinstance(e, UnknownFunction)
    assert e.proto_err.location[-1].node_idx == foo.idx
    sig = Signature(
        root=Namespace(
            functions={"foo": _foo_func},
        ),
    )
    tg = infer_graph_types(tg, sig)
    out_type = tg.get_edge(NodePort(foo, "res"), NodePort(tg.output, "out")).type_
    assert out_type == PairType(IntType(), IntType())


@pytest.mark.skip_typecheck
@pytest.mark.asyncio
async def test_infer_graph_types_with_sig(client: RuntimeClient):
    # client is only used for signatures of builtins etc.
    sigs = await client.get_signature()

    tg = TierkreisGraph()
    mkp = tg.add_func(
        "make_pair",
        first=tg.input["in"],
        second=tg.add_const(3),
    )
    tg.set_outputs(val=mkp["pair"])

    # updates the graph wrapped by tg inplace
    tg._graph = infer_graph_types(tg, sigs)._graph
    in_type = tg.get_edge(NodePort(tg.input, "in"), NodePort(mkp, "first")).type_
    assert isinstance(in_type, VarType)
    out_type = tg.get_edge(NodePort(mkp, "pair"), NodePort(tg.output, "val")).type_
    assert out_type == PairType(in_type, IntType())


@pytest.mark.skip_typecheck
@pytest.mark.asyncio
async def test_infer_graph_types_with_inputs(
    client: RuntimeClient, idpy_graph: TierkreisGraph
):
    ns = Namespace(
        functions={"foo": _foo_func},
        subspaces={
            "python_nodes": (await client.get_signature()).root.subspaces[
                "python_nodes"
            ]
        },
    )
    funcs = Signature(
        root=ns,
    )
    tg = TierkreisGraph()
    foo = tg.add_func("foo", value=tg.input["inp"])
    tg.set_outputs(out=foo["res"])

    tg2 = deepcopy(tg)

    inputs: StructValue = StructValue({"inp": FloatValue(3.14)})
    tg, inputs_ = infer_graph_types(tg, funcs, inputs)
    assert inputs_ == inputs
    out_type = tg.get_edge(NodePort(foo, "res"), NodePort(tg.output, "out")).type_
    assert out_type == PairType(FloatType(), IntType())

    graph_inputs: StructValue = StructValue({"inp": GraphValue(idpy_graph)})
    with pytest.raises(TierkreisTypeErrors, match=r"Pair\[Float, Int\] | Float") as err:
        # Pass an argument inconsistent with the annotations now on tg
        infer_graph_types(tg, funcs, graph_inputs)

    assert len(err.value) == 1
    e = err.value.errors[0]
    assert isinstance(e, UnifyError)
    # TODO location sometimes
    assert betterproto.which_one_of(e.proto_err.location[-1], "location")[0] in (
        "input",
        "root",
    )
    # deep copy (above) has no annotations yet, so ok
    tg2, inputs_ = infer_graph_types(tg2, funcs, graph_inputs)
    out_type = tg2.get_edge(NodePort(foo, "res"), NodePort(tg2.output, "out")).type_
    assert isinstance(out_type, PairType) and out_type.second == IntType()
    assert isinstance(out_type.first, GraphType)
    argtypes, restypes = out_type.first.inputs, out_type.first.outputs
    assert argtypes.rest is restypes.rest is None
    assert len(argtypes.content) == len(restypes.content) == 1
    assert argtypes.content["id_in"] == restypes.content["id_out"]
    assert isinstance(argtypes.content["id_in"], VarType)


@pytest.mark.asyncio
async def test_deep_type_err(client: RuntimeClient, bi: "BuilderNS"):
    @graph()
    def ifelse() -> Output:
        with IfElse(Const(True)):
            with If():
                bi.iadd(Const(1), Const(1.0))
            with Else():
                pass
        return Output()

    if not client.can_type_check:
        pytest.skip()
    with pytest.raises(
        TierkreisTypeErrors,
        match=r"In: NodeRef\(3, Const\(Graph\(if\)\)\)\/NodeRef\(4, Function\(iadd\)\)"
        + r"| NodeRef\(3, Const\(Graph\(if\)\)\)\/NodeRef\(3, Const\(1.0\)\)",
    ):
        await client.type_check_graph(ifelse)


def _extract_dbg_info(errstrings: str) -> Iterator[tuple[str, int]]:
    for errstr in errstrings.split("\n\n" + ("─" * 80) + "\n\n"):
        lines = errstr.splitlines()

        lines = lines[lines.index("Debug Info:") + 1 :]
        for line in lines:
            _, fname, linno = line.split(":", 3)
            yield fname.strip(), int(linno.strip())


@pytest.mark.skip_typecheck
@pytest.mark.asyncio
async def test_builder_debug(bi: "BuilderNS", sig: Signature):
    # test debug string is correctly found
    # checks that "ERROR" preceeds the location of reported errors
    def _bad_nod():
        # ERROR
        return bi.iadd(Const(1), Const(1.0))

    @lazy_graph(type_check_sig=sig)
    def ifelse() -> Output:
        # ERROR
        with IfElse(Const(True)):
            with If():
                _bad_nod()
            with Else():
                # ERROR
                pass
        return Output()

    @lazy_graph(type_check_sig=sig)
    def lop() -> Output:
        @loop()
        def lopdef(x):
            # ERROR
            with IfElse(Const(True)):
                with If():
                    # ERROR
                    Continue(bi.iadd(Const(1), Const(1.0)))
                with Else():
                    # ERROR
                    Break(Const(1))
            return Output()

        # ERROR
        _y = lopdef(Const(1))

        return Output()

    @lazy_graph(type_check_sig=sig)
    def clos1() -> Output:
        # ERROR
        y = Const(1.0)

        @closure()
        # ERROR
        def clos_def(x):
            # ERROR
            bi.iadd(x, y)
            return Output()

        return Output()

    @lazy_graph(type_check_sig=sig)
    def clos2() -> Output:
        @closure()
        # ERROR
        def clos_def(x):
            # ERROR
            bi.iadd(x, Const(1))
            return Output()

        # ERROR
        _y = clos_def(Const(1.0))

        return Output()

    @lazy_graph(type_check_sig=sig)
    def scop(x: ValueSource) -> Output:
        # ERROR
        with Scope():
            # ERROR
            with Scope():
                # ERROR
                x2 = _bad_nod()
        return Output(x2)

    for g in (ifelse, lop, clos1, clos2, scop):
        with pytest.raises(TierkreisTypeErrors) as e:
            _ = g.graph
        if (
            sys.platform.startswith("win")
            or int(platform.python_version_tuple()[1]) > 10
        ):
            # string parsing doesn't work on windows or 3.11
            continue
        with open(__file__) as f:
            self_lines = f.readlines()
        for fil, linno in _extract_dbg_info(str(e.value)):
            assert fil == __file__
            # minus 1 for 1 indexed line numbers
            assert "# ERROR" in self_lines[linno - 2]



"""Visualise TierkreisGraph using graphviz."""

import textwrap
from typing import Iterable, NewType, Optional, Tuple, cast

import graphviz as gv

from tierkreis.core.tierkreis_graph import (
    BoxNode,
    ConstNode,
    FunctionNode,
    GraphValue,
    InputNode,
    Location,
    MatchNode,
    OutputNode,
    TagNode,
    TierkreisGraph,
    TierkreisNode,
)

# old palettte: https://colorhunt.co/palette/343a407952b3ffc107e1e8eb
# _COLOURS = {
#     "background": "white",
#     "node": "#7952B3",
#     "edge": "#FFC107",
#     "dark": "#343A40",
#     "const": "#7c55b4",
#     "discard": "#ff8888",
#     "node_border": "#9d80c7",
#     "port_border": "#ffd966",
# }

# ZX colours
# _COLOURS = {
#     "background": "white",
#     "node": "#629DD1",
#     "edge": "#297FD5",
#     "dark": "#112D4E",
#     "const": "#a1eea1",
#     "discard": "#ff8888",
#     "node_border": "#D8F8D8",
#     "port_border": "#E8A5A5",
# }

# Conference talk colours
_COLOURS = {
    "background": "white",
    "node": "#ACCBF9",
    "edge": "#1CADE4",
    "dark": "black",
    "const": "#77CEEF",
    "discard": "#ff8888",
    "node_border": "white",
    "port_border": "#1CADE4",
}


_FONTFACE = "monospace"

_HTML_LABEL_TEMPLATE = """
<TABLE BORDER="{border_width}" CELLBORDER="0" CELLSPACING="1" CELLPADDING="1" BGCOLOR="{node_back_color}" COLOR="{border_colour}">

{inputs_row}

    <TR>
        <TD>
            <TABLE BORDER="0" CELLBORDER="0">
                <TR>
                    <TD><FONT POINT-SIZE="{fontsize}" FACE="{fontface}" COLOR="{label_color}"><B>{node_label}</B></FONT></TD>
                </TR>
            </TABLE>
        </TD>
    </TR>

{outputs_row}

</TABLE>
"""


def _format_html_label(**kwargs):
    _HTML_LABEL_DEFAULTS = {
        "label_color": _COLOURS["dark"],
        "node_back_color": _COLOURS["node"],
        "inputs_row": "",
        "outputs_row": "",
        "border_colour": _COLOURS["port_border"],
        "border_width": "1",
        "fontface": _FONTFACE,
        "fontsize": 11.0,
    }
    return _HTML_LABEL_TEMPLATE.format(**{**_HTML_LABEL_DEFAULTS, **kwargs})


_HTML_PORTS_ROW_TEMPLATE = """
    <TR>
        <TD>
            <TABLE BORDER="0" CELLBORDER="0" CELLSPACING="3" CELLPADDING="2">
                <TR>
                    {port_cells}
                </TR>
            </TABLE>
        </TD>
    </TR>
"""

_HTML_PORT_TEMPLATE = (
    '<TD BGCOLOR="{back_colour}" COLOR="{border_colour}"'
    ' PORT="{port_id}" BORDER="{border_width}">'
    '<FONT POINT-SIZE="10.0" FACE="{fontface}" COLOR="{font_colour}">{port}</FONT></TD>'
)

_INPUT_PREFIX = "in."
_OUTPUT_PREFIX = "out."


def _html_ports(ports: Iterable[str], id_prefix: str) -> str:
    return _HTML_PORTS_ROW_TEMPLATE.format(
        port_cells="".join(
            _HTML_PORT_TEMPLATE.format(
                port=port,
                # differentiate input and output node identifiers
                # with a prefix
                port_id=id_prefix + port,
                back_colour=_COLOURS["background"],
                font_colour=_COLOURS["dark"],
                border_width="1",
                border_colour=_COLOURS["port_border"],
                fontface=_FONTFACE,
            )
            for port in ports
        )
    )


def _trim_str(instr: str, max_len: int = 10) -> str:
    return textwrap.shorten(instr, width=max_len, placeholder="...")


def _thunk_name(thunk: TierkreisGraph) -> str:
    name = "thunk"
    if thunk.name:
        name += f"\n<BR/><I>{thunk.name}</I>"

    return name


def _print_location(loc: Location) -> str:
    if loc.location == []:
        return ""
    return "@" + "/".join(loc.location)


def _node_id(prefix: str, idx: int, sep: str = "//") -> str:
    return f"{prefix}{sep}{idx}"


def _node_features(node: TierkreisNode) -> Tuple[str, str]:
    """Calculate node label (first) and colour (second)."""

    fillcolor = _COLOURS["node"]
    if isinstance(node, FunctionNode):
        f_name = node.function_name
        node_label = str(f_name)
    elif isinstance(node, ConstNode):
        fillcolor = _COLOURS["const"]
        value = node.value
        if isinstance(value, GraphValue):
            fillcolor = _COLOURS["edge"]
            node_label = _thunk_name(value.value)
        else:
            const_str = node.value.viz_str()
            node_label = _trim_str(const_str, 15)
    elif isinstance(node, MatchNode):
        node_label = "Match"
    elif isinstance(node, TagNode):
        node_label = "Tag: " + node.tag_name
    elif isinstance(node, BoxNode):
        node_label = f"Box{_print_location(node.location)}"
        name = cast(BoxNode, node).graph.name
        if name:
            node_label += f": {name}"
    elif isinstance(node, (InputNode, OutputNode)):
        # effectively only leave the ports visible
        fillcolor = _COLOURS["background"]
        node_label = " "
    else:
        node_label = ""

    return node_label, fillcolor


def tierkreis_to_graphviz(
    tk_graph: TierkreisGraph,
    initial_graph: Optional[gv.Digraph] = None,
    prefix: str = "",
    unbox_level: int = 0,
    merge_copies: bool = True,
    unbox_graph_names: Optional[set[str]] = None,
) -> gv.Digraph:
    """Return a visual representation of the TierkreisGraph as a graphviz object.

    Returns:
        graphviz.DiGraph: Representation of the TierkreisGraph
    """
    if merge_copies:
        tk_graph = _merge_copies(tk_graph)
    gv_graph = (
        gv.Digraph(
            tk_graph.name or "Tierkreis",
            strict=False,  # stops multiple shared edges being merged
        )
        if initial_graph is None
        else initial_graph
    )
    graph_atrr = {
        "rankdir": "",
        "ranksep": "0.1",
        "nodesep": "0.15",
        "margin": "0",
        "bgcolor": _COLOURS["background"],
    }
    if tk_graph.name:
        graph_label = _format_html_label(
            node_label=tk_graph.name,
            fontsize=11.0,
            border_width=2.0,
            node_back_color=_COLOURS["background"],
        )
        graph_atrr["label"] = f"<{graph_label}>"
        graph_atrr["labelloc"] = "t"
    gv_graph.attr(**graph_atrr)

    unboxed_nodes = set()
    unthunked_nodes = set()
    discard_nodes = set()
    copy_nodes = set()
    no_outport_nodes = set()
    for node_idx, node in enumerate(tk_graph.nodes()):
        node_identifier = _node_id(prefix, node_idx)
        if node.is_discard_node():
            gv_graph.node(
                node_identifier,
                label="",
                shape="point",
                color=_COLOURS["discard"],
                width="0.1",
            )
            discard_nodes.add(node_identifier)
            continue
        if node.is_copy_node():
            gv_graph.node(
                node_identifier,
                label="",
                shape="point",
                color=_COLOURS["edge"],
                width="0.1",
            )
            copy_nodes.add(node_identifier)
            continue

        node_label, fillcolor = _node_features(node)

        # node is a table
        # first row is a single cell containing a single row table of inputs
        # second row is table containing single cell of node_label
        # third row is single cell containing a single row table of outputs

        in_ports = [edge.target.port for edge in tk_graph.in_edges(node_idx)]
        out_ports = [edge.source.port for edge in tk_graph.out_edges(node_idx)]
        subgraph = None
        isbox = False
        isgraphconst = False
        if isinstance(node, BoxNode):
            isbox = True
            subgraph = node.graph
        if isinstance(node, ConstNode):
            val = node.value
            if isinstance(val, GraphValue):
                subgraph = val.value
                isgraphconst = True
        if (
            unbox_level > 0
            and subgraph is not None
            and (not unbox_graph_names or subgraph.name in unbox_graph_names)
        ):
            cluster_name = "cluster" + node_identifier
            if isbox:
                unboxed_nodes.add(node_identifier)
                subgraph = cast(BoxNode, node).graph
                out_ports = []
            else:
                unthunked_nodes.add(node_identifier)
                subgraph = cast(GraphValue, cast(ConstNode, node).value).value

            ctx_mgr = gv_graph.subgraph(name=cluster_name)
            assert ctx_mgr is not None
            with ctx_mgr as c:
                tierkreis_to_graphviz(
                    subgraph,
                    initial_graph=c,
                    prefix=node_identifier,
                    unbox_level=unbox_level - 1,
                )

                if isgraphconst:
                    html_label = _format_html_label(
                        node_back_color=_COLOURS["edge"],
                        node_label=_thunk_name(subgraph),
                        border_colour=_COLOURS["port_border"],
                    )
                    c.node(
                        node_identifier + "thunk",
                        shape="plain",
                        label=f"<{html_label}>",
                    )
                    c.attr(label="")
                else:
                    html_label = _format_html_label(
                        node_back_color=fillcolor,
                        border_colour=_COLOURS["node_border"],
                        node_label=(node_label) if isbox else "Thunk",
                        outputs_row=_html_ports(out_ports, _OUTPUT_PREFIX)
                        if out_ports
                        else "",
                    )
                    c.attr(label=f"<{html_label}>")
                c.attr(
                    margin="10",
                    style="dashed",
                    color=(_COLOURS["const"], _COLOURS["edge"])[unbox_level % 2],
                )
            continue

        if isinstance(node, ConstNode):
            # unnecessary "value" port for constants
            no_outport_nodes.add(node_identifier)
            out_ports = []

        html_label = _format_html_label(
            node_back_color=fillcolor,
            node_label=node_label,
            inputs_row=_html_ports(in_ports, _INPUT_PREFIX) if in_ports else "",
            outputs_row=_html_ports(out_ports, _OUTPUT_PREFIX) if out_ports else "",
            border_colour=_COLOURS["background"]
            if fillcolor == _COLOURS["background"]
            else _COLOURS["node_border"],
        )
        gv_graph.node(
            node_identifier,
            label=f"<{html_label}>",
            shape="plain",
        )

    edge_attr = {
        "penwidth": "1.5",
        "arrowhead": "none",
        "arrowsize": "1.0",
        "fontname": _FONTFACE,
        "fontsize": "9",
        "color": _COLOURS["edge"],
        "fontcolor": "black",
    }
    for edge in tk_graph.edges():
        src_node = _node_id(prefix, edge.source.node_ref.idx)
        tgt_node = _node_id(prefix, edge.target.node_ref.idx)

        if src_node in unboxed_nodes:
            src_node = _node_id(src_node, TierkreisGraph.output_node_idx)
            # box output node only has input ports
            outport_str = (
                ""
                if src_node in no_outport_nodes
                else f":{_INPUT_PREFIX}{edge.source.port}"
            )
            src_nodeport = f"{src_node}{outport_str}"
        elif src_node in unthunked_nodes:
            src_nodeport = src_node + "thunk"
        elif src_node in copy_nodes:
            src_nodeport = src_node
        else:
            outport_str = (
                ""
                if src_node in no_outport_nodes
                else f":{_OUTPUT_PREFIX}{edge.source.port}"
            )
            src_nodeport = f"{src_node}{outport_str}"

        if tgt_node in unboxed_nodes:
            tgt_node = _node_id(tgt_node, TierkreisGraph.input_node_idx)
            # box input node only has output ports
            tgt_nodeport = f"{tgt_node}:{_OUTPUT_PREFIX}{edge.target.port}"
        elif tgt_node in discard_nodes or tgt_node in copy_nodes:
            # discard/copy nodes don't have ports (not HTML labels)
            tgt_nodeport = tgt_node
        else:
            tgt_nodeport = f"{tgt_node}:{_INPUT_PREFIX}{edge.target.port}"

        if src_node in copy_nodes or edge.type_ is None:
            # copy edge output types can be seen from the input
            edge_label = ""
        else:
            edge_label = textwrap.fill(
                str(edge.type_), 20, fix_sentence_endings=True, break_long_words=False
            )
        gv_graph.edge(
            src_nodeport,
            tgt_nodeport,
            label=edge_label,
            **edge_attr,
        )

    return gv_graph


def render_graph(
    graph: TierkreisGraph, filename: str, format_st: str, **kwargs
) -> None:
    """Use graphviz to render a graph visualisation to file

    Args:
        graph: Graph to render
        filename: Filename root to write render to
        format_st: Format string, e.g. "png", refer to Graphviz
            render
    documentation for full list.
    """
    gv_graph = tierkreis_to_graphviz(graph, **kwargs)

    gv_graph.render(filename, format=format_st)


# Merging copies produces invalid graphs
_CopyMergedGraph = NewType("_CopyMergedGraph", TierkreisGraph)


def _merge_copies(g: TierkreisGraph) -> _CopyMergedGraph:
    """Merge adjacent copy nodes - adds extra ports to copy nodes so won't pass
    type check."""
    candidates = {idx for idx, node in enumerate(g.nodes()) if node.is_copy_node()}
    while candidates:
        node_name = candidates.pop()
        copy_children = (
            n.idx
            for e in g.out_edges(node_name)
            if g[(n := e.target.node_ref)].is_copy_node()
        )

        if (copy_child := next(copy_children, None)) is None:
            continue

        # this node is going to merge with child - check it again
        candidates.add(node_name)

        eds = list(g.out_edges(copy_child))

        # remove child and all connected edges
        g._graph.remove_node(copy_child)
        # remove from candidates if still a candidate
        if copy_child in candidates:
            candidates.remove(copy_child)

        for e in eds:
            # source port is not valid - this graph will not type check
            g._graph.add_edge(
                node_name, e.target.node_ref.idx, ("", e.target.port), type=e.type_
            )

    return _CopyMergedGraph(g)
